"""Types of the objects contained in the OSEF stream."""

# Standard imports
from typing import Dict, List, NamedTuple, Union

# Third party imports
import numpy as np

# Project imports
from osef.spec import _formats, _packers
from osef.spec.osef_types import OsefKeys, OsefTypes


class InternalNodeInfo(NamedTuple):
    """Internal node info, used for parent node.

    For now, only list or dict are defined.
    :param node_type: Type of the node.
    """

    node_type: Union[Dict, List]


class TypeInfo(NamedTuple):
    """Define the type info on an OSEF type.

    :param name: OSEF key.
    :param node_info: Information on the data type.
    """

    name: str
    node_info: Union[InternalNodeInfo, _packers.PackerBase]


outsight_types = {
    OsefTypes.AUGMENTED_CLOUD: TypeInfo(
        OsefKeys.AUGMENTED_CLOUD, InternalNodeInfo(dict)
    ),
    OsefTypes.NUMBER_OF_POINTS: TypeInfo(
        OsefKeys.NUMBER_OF_POINTS,
        _packers.ValuePacker(_formats.BinaryValueFormat.NB_POINTS),
    ),
    OsefTypes.SPHERICAL_COORDINATES: TypeInfo(
        OsefKeys.SPHERICAL_COORDINATES,
        _packers.StructuredArrayPacker(_formats.SphericalCoodinatesDType),
    ),
    OsefTypes.REFLECTIVITIES: TypeInfo(
        OsefKeys.REFLECTIVITIES,
        _packers.ArrayPacker(np.dtype(np.uint8)),
    ),
    OsefTypes._BACKGROUND_FLAGS: TypeInfo(
        OsefKeys._BACKGROUND_FLAGS,
        _packers.ArrayPacker(np.dtype(np.bool_)),
    ),
    OsefTypes.CARTESIAN_COORDINATES: TypeInfo(
        OsefKeys.CARTESIAN_COORDINATES,
        _packers.ArrayPacker(np.dtype(np.float32), 3),
    ),
    OsefTypes._BGR_COLORS: TypeInfo(
        OsefKeys._BGR_COLORS,
        _packers.BytesPacker(),
    ),
    OsefTypes._OBJECT_DETECTION_FRAME: TypeInfo(
        OsefKeys._OBJECT_DETECTION_FRAME, InternalNodeInfo(dict)
    ),
    OsefTypes._IMAGE_DIMENSION: TypeInfo(
        OsefKeys._IMAGE_DIMENSION,
        _packers.DictPacker(*_formats.ImageDimensionDictFormat),
    ),
    OsefTypes.NUMBER_OF_OBJECTS: TypeInfo(
        OsefKeys.NUMBER_OF_OBJECTS,
        _packers.ValuePacker(_formats.BinaryValueFormat.NB_OBJECTS),
    ),
    OsefTypes._CLOUD_FRAME: TypeInfo(OsefKeys._CLOUD_FRAME, InternalNodeInfo(dict)),
    OsefTypes.TIMESTAMP_MICROSECOND: TypeInfo(
        OsefKeys.TIMESTAMP_MICROSECOND,
        _packers.TimestampMicroPacker(),
    ),
    OsefTypes._AZIMUTHS_COLUMN: TypeInfo(
        OsefKeys._AZIMUTHS_COLUMN,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.NUMBER_OF_LAYERS: TypeInfo(
        OsefKeys.NUMBER_OF_LAYERS,
        _packers.ValuePacker(_formats.BinaryValueFormat.NB_LAYERS),
    ),
    OsefTypes._CLOUD_PROCESSING: TypeInfo(
        OsefKeys._CLOUD_PROCESSING,
        _packers.ProcessingBitfieldPacker(),
    ),
    OsefTypes._RANGE_AZIMUTH: TypeInfo(
        OsefKeys._RANGE_AZIMUTH,
        _packers.DictPacker(*_formats.RangeAzimuthDictFormat),
    ),
    OsefTypes._BOUNDING_BOXES_ARRAY: TypeInfo(
        OsefKeys._BOUNDING_BOXES_ARRAY,
        _packers.StructuredArrayPacker(
            _formats.BoundingBoxesDType,
        ),
    ),
    OsefTypes.CLASS_ID_ARRAY: TypeInfo(
        OsefKeys.CLASS_ID_ARRAY,
        _packers.ArrayPacker(np.dtype(np.int32)),
    ),
    OsefTypes._CONFIDENCE_ARRAY: TypeInfo(
        OsefKeys._CONFIDENCE_ARRAY,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.TIMESTAMPED_DATA: TypeInfo(
        OsefKeys.TIMESTAMPED_DATA, InternalNodeInfo(dict)
    ),
    OsefTypes._PERCEPT: TypeInfo(
        OsefKeys._PERCEPT,
        _packers.ArrayPacker(np.dtype(np.int32)),
    ),
    OsefTypes._BGR_IMAGE_FRAME: TypeInfo(
        OsefKeys._BGR_IMAGE_FRAME, InternalNodeInfo(dict)
    ),
    OsefTypes.POSE: TypeInfo(
        OsefKeys.POSE,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.SCAN_FRAME: TypeInfo(OsefKeys.SCAN_FRAME, InternalNodeInfo(dict)),
    OsefTypes.TRACKED_OBJECTS: TypeInfo(
        OsefKeys.TRACKED_OBJECTS, InternalNodeInfo(dict)
    ),
    OsefTypes.BBOX_SIZES: TypeInfo(
        OsefKeys.BBOX_SIZES,
        _packers.ArrayPacker(np.dtype(np.float32), 3),
    ),
    OsefTypes.SPEED_VECTORS: TypeInfo(
        OsefKeys.SPEED_VECTORS,
        _packers.ArrayPacker(np.dtype(np.float32), 3),
    ),
    OsefTypes.POSE_ARRAY: TypeInfo(
        OsefKeys.POSE_ARRAY,
        _packers.ArrayPacker(*_formats.PoseArrayFormat),
    ),
    OsefTypes.OBJECT_ID: TypeInfo(
        OsefKeys.OBJECT_ID,
        _packers.ArrayPacker(np.dtype(np.ulonglong)),
    ),
    OsefTypes.CARTESIAN_COORDINATES_4F: TypeInfo(
        OsefKeys.CARTESIAN_COORDINATES_4F,
        _packers.StructuredArrayPacker(_formats.CartesianCoordinates4fDType),
    ),
    OsefTypes.SPHERICAL_COORDINATES_4F: TypeInfo(
        OsefKeys.SPHERICAL_COORDINATES_4F,
        _packers.StructuredArrayPacker(_formats.SphericalCoodinates4fDType),
    ),
    OsefTypes.ZONES_DEF: TypeInfo(OsefKeys.ZONES_DEF, InternalNodeInfo(list)),
    OsefTypes.ZONE: TypeInfo(OsefKeys.ZONE, InternalNodeInfo(dict)),
    OsefTypes.ZONE_VERTICES: TypeInfo(
        OsefKeys.ZONE_VERTICES,
        _packers.ArrayPacker(np.dtype(np.float32), 2),
    ),
    OsefTypes.ZONE_NAME: TypeInfo(
        OsefKeys.ZONE_NAME,
        _packers.StringPacker(),
    ),
    OsefTypes._ZONE_UUID: TypeInfo(
        OsefKeys._ZONE_UUID,
        _packers.UuidPacker(),
    ),
    OsefTypes.ZONES_OBJECTS_BINDING: TypeInfo(
        OsefKeys.ZONES_OBJECTS_BINDING,
        _packers.StructuredArrayPacker(_formats.ZoneObjectBindingDType),
    ),
    OsefTypes.OBJECT_PROPERTIES: TypeInfo(
        OsefKeys.OBJECT_PROPERTIES,
        _packers.ObjectPropertiesPacker(),
    ),
    OsefTypes._IMU_PACKET: TypeInfo(
        OsefKeys._IMU_PACKET,
        _packers.ImuPacker(),
    ),
    OsefTypes._TIMESTAMP_LIDAR_VELODYNE: TypeInfo(
        OsefKeys._TIMESTAMP_LIDAR_VELODYNE,
        _packers.DictPacker(*_formats.TimestampLidarVelodyneDictFormat),
    ),
    OsefTypes.POSE_RELATIVE: TypeInfo(
        OsefKeys.POSE_RELATIVE,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._GRAVITY: TypeInfo(
        OsefKeys._GRAVITY,
        _packers.DictPacker(*_formats.GravityDictFormat),
    ),
    OsefTypes.EGO_MOTION: TypeInfo(OsefKeys.EGO_MOTION, InternalNodeInfo(dict)),
    OsefTypes._PREDICTED_POSITION: TypeInfo(
        OsefKeys._PREDICTED_POSITION,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.GEOGRAPHIC_POSE: TypeInfo(
        OsefKeys.GEOGRAPHIC_POSE,
        _packers.StructuredArrayPacker(_formats.GeographicPoseDType),
    ),
    OsefTypes.OBJECT_ID_32_BITS: TypeInfo(
        OsefKeys.OBJECT_ID_32_BITS,
        _packers.ArrayPacker(np.dtype(np.uint32)),
    ),
    OsefTypes.ZONES_OBJECTS_BINDING_32_BITS: TypeInfo(
        OsefKeys.ZONES_OBJECTS_BINDING_32_BITS,
        _packers.StructuredArrayPacker(_formats.ZoneObjectBinding32bDType),
    ),
    OsefTypes._BACKGROUND_BITS: TypeInfo(
        OsefKeys._BACKGROUND_BITS,
        _packers.BoolBitfieldPacker(),
    ),
    OsefTypes._GROUND_PLANE_BITS: TypeInfo(
        OsefKeys._GROUND_PLANE_BITS,
        _packers.BoolBitfieldPacker(),
    ),
    OsefTypes._AZIMUTHS: TypeInfo(
        OsefKeys._AZIMUTHS,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._ELEVATIONS: TypeInfo(
        OsefKeys._ELEVATIONS,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._DISTANCES: TypeInfo(
        OsefKeys._DISTANCES,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._LIDAR_MODEL: TypeInfo(
        OsefKeys._LIDAR_MODEL,
        _packers.ValuePacker(_formats.BinaryValueFormat.LIDAR_MODEL),
    ),
    OsefTypes.SLAM_POSE_ARRAY: TypeInfo(
        OsefKeys.SLAM_POSE_ARRAY,
        _packers.ArrayPacker(*_formats.PoseArrayFormat),
    ),
    OsefTypes.ZONE_VERTICAL_LIMITS: TypeInfo(
        OsefKeys.ZONE_VERTICAL_LIMITS,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.GEOGRAPHIC_POSE_PRECISE: TypeInfo(
        OsefKeys.GEOGRAPHIC_POSE_PRECISE,
        _packers.StructuredArrayPacker(_formats.GeographicPosePreciseDType),
    ),
    OsefTypes._ROAD_MARKINGS_BITS: TypeInfo(
        OsefKeys._ROAD_MARKINGS_BITS,
        _packers.BoolBitfieldPacker(),
    ),
    OsefTypes.SMOOTHED_POSE: TypeInfo(
        OsefKeys.SMOOTHED_POSE,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._HEIGHT_MAP: TypeInfo(OsefKeys._HEIGHT_MAP, InternalNodeInfo(dict)),
    OsefTypes._HEIGHT_MAP_POINTS: TypeInfo(
        OsefKeys._HEIGHT_MAP_POINTS,
        _packers.ArrayPacker(np.dtype(np.float32), 3),
    ),
    OsefTypes.DIVERGENCE_INDICATOR: TypeInfo(
        OsefKeys.DIVERGENCE_INDICATOR,
        _packers.ValuePacker(_formats.BinaryValueFormat.DIVERGENCE),
    ),
    OsefTypes._CARLA_TAG_ARRAY: TypeInfo(
        OsefKeys._CARLA_TAG_ARRAY,
        _packers.ArrayPacker(np.dtype(np.uint32)),
    ),
    OsefTypes._BACKGROUND_SCENE_PARAMS: TypeInfo(
        OsefKeys._BACKGROUND_SCENE_PARAMS, InternalNodeInfo(dict)
    ),
    OsefTypes._BACKGROUND_SCENE_PARAMS_GENERAL: TypeInfo(
        OsefKeys._BACKGROUND_SCENE_PARAMS_GENERAL,
        _packers.StructuredArrayPacker(_formats.BackgroundSceneParamsDType),
    ),
    OsefTypes._BACKGROUND_SCENE_PARAMS_ELEVATIONS: TypeInfo(
        OsefKeys._BACKGROUND_SCENE_PARAMS_ELEVATIONS,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._BACKGROUND_SCENE_FRAGMENT: TypeInfo(
        OsefKeys._BACKGROUND_SCENE_FRAGMENT, InternalNodeInfo(dict)
    ),
    OsefTypes._BACKGROUND_SCENE_FRAGMENT_INFO: TypeInfo(
        OsefKeys._BACKGROUND_SCENE_FRAGMENT_INFO,
        _packers.StructuredArrayPacker(_formats.BackgroundSceneFragmentInfoDType),
    ),
    OsefTypes._BACKGROUND_SCENE_FRAGMENT_DISTANCES: TypeInfo(
        OsefKeys._BACKGROUND_SCENE_FRAGMENT_DISTANCES,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.GEOGRAPHIC_POSE_ARRAY: TypeInfo(
        OsefKeys.GEOGRAPHIC_POSE_ARRAY,
        _packers.StructuredArrayPacker(_formats.GeographicPosePreciseDType),
    ),
    OsefTypes.GEOGRAPHIC_SPEED: TypeInfo(
        OsefKeys.GEOGRAPHIC_SPEED,
        _packers.StructuredArrayPacker(_formats.GeographicSpeedDType),
    ),
    OsefTypes.GEOGRAPHIC_SPEED_ARRAY: TypeInfo(
        OsefKeys.GEOGRAPHIC_SPEED_ARRAY,
        _packers.StructuredArrayPacker(_formats.GeographicSpeedDType),
    ),
    OsefTypes._INSTANTANEOUS_TRANSLATION_SPEED: TypeInfo(
        OsefKeys._INSTANTANEOUS_TRANSLATION_SPEED,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._INSTANTANEOUS_ROTATION_SPEED: TypeInfo(
        OsefKeys._INSTANTANEOUS_ROTATION_SPEED,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._FILTERED_TRANSLATION_SPEED: TypeInfo(
        OsefKeys._FILTERED_TRANSLATION_SPEED,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._FILTERED_ROTATION_SPEED: TypeInfo(
        OsefKeys._FILTERED_ROTATION_SPEED,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.REFERENCE_MAP_BITS: TypeInfo(
        OsefKeys.REFERENCE_MAP_BITS,
        _packers.BoolBitfieldPacker(),
    ),
    OsefTypes._CARTESIAN_COVARIANCE: TypeInfo(
        OsefKeys._CARTESIAN_COVARIANCE, _packers.ArrayPacker(np.dtype(np.float32))
    ),
    OsefTypes._CYLINDRICAL_COVARIANCE: TypeInfo(
        OsefKeys._CYLINDRICAL_COVARIANCE,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes.COORDINATES_REFERENCE_SYSTEM: TypeInfo(
        OsefKeys.COORDINATES_REFERENCE_SYSTEM,
        _packers.ValuePacker(_formats.BinaryValueFormat.COORDINATES_REFERENCE_SYSTEM),
    ),
    OsefTypes.START_TIMESTAMP: TypeInfo(
        OsefKeys.START_TIMESTAMP,
        _packers.TimestampNanoPacker(),
    ),
    OsefTypes.END_TIMESTAMP: TypeInfo(
        OsefKeys.END_TIMESTAMP,
        _packers.TimestampNanoPacker(),
    ),
    OsefTypes.TIME_DOMAIN: TypeInfo(
        OsefKeys.TIME_DOMAIN,
        _packers.ValuePacker(_formats.BinaryValueFormat.TIME_DOMAIN),
    ),
    OsefTypes.TIME_INTERVAL: TypeInfo(OsefKeys.TIME_INTERVAL, InternalNodeInfo(dict)),
    OsefTypes.CLASSIFICATION_CONFIDENCE: TypeInfo(
        OsefKeys.CLASSIFICATION_CONFIDENCE,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._CLASS_PROBABILITIES_ARRAY: TypeInfo(
        OsefKeys._CLASS_PROBABILITIES_ARRAY,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._CLASSIFIER_DEF: TypeInfo(
        OsefKeys._CLASSIFIER_DEF, InternalNodeInfo(dict)
    ),
    OsefTypes._ENABLED_CLASS_IDS_ARRAY: TypeInfo(
        OsefKeys._ENABLED_CLASS_IDS_ARRAY,
        _packers.ArrayPacker(np.dtype(np.uint32)),
    ),
    OsefTypes._REFLECTION_FEATURES_ARRAY: TypeInfo(
        OsefKeys._REFLECTION_FEATURES_ARRAY,
        _packers.ArrayPacker(np.dtype(np.float32)),
    ),
    OsefTypes._REFLECTION_BITS: TypeInfo(
        OsefKeys._REFLECTION_BITS,
        _packers.BoolBitfieldPacker(),
    ),
}


def get_type_info_by_id(type_code: int):
    """Get TypeInfo for a given type code.

    :param type_code: Int value in OsefTypes
    :return:
    """
    if type_code in outsight_types:
        return outsight_types[type_code]

    return TypeInfo(f"Unknown type ({type_code})", _packers.PackerBase())


def get_type_info_by_key(type_name: str) -> TypeInfo:
    """Get TypeInfo for a given key/name.

    :param type_name: Int value in OsefTypes
    :return:
    """
    for value in outsight_types.values():
        if value.name == type_name:
            return value
    return TypeInfo(f"Unknown type ({type_name})", _packers.PackerBase())


def get_type_by_key(type_name: str) -> OsefTypes:
    """Get type index for a given key/name.

    :param type_name: Int value in OsefTypes
    :return:
    """
    try:
        return OsefTypes[OsefKeys(type_name).name]
    except KeyError as key_error:
        raise KeyError(f"No type found for type_name={type_name}") from key_error



"""Functions to pack data to osef format"""
import traceback
from struct import Struct
from typing import Any, Union

from osef.spec import _types
from osef._logger import osef_logger
from osef.spec.constants import _Tlv, _TreeNode, _STRUCT_FORMAT
from osef.spec.osef_types import OsefKeys
from osef.parsing._parser_common import _align_size
from osef.parsing._lazy_parser import _LazyDict, _LazyList, BinaryLeaf
from osef.spec import _packers


# -- Public functions --
def pack(frame: dict) -> bytes:
    """Encode an osef frame content to a TLV

    :param frame: osef frame dict to be packed in the OSEF format (TLV)
    :raise KeyError: Raises exception on unknown type
    :return: bytes tlv
    """
    tree = _pack_to_tree(
        _types.OsefTypes.TIMESTAMPED_DATA, frame[OsefKeys.TIMESTAMPED_DATA.value]
    )
    return _tree_to_bin(tree)


def _encode_tlv(tlv: _Tlv) -> bytes:
    """Encode a single TLV to bytes sequence"""
    return Struct(_STRUCT_FORMAT % _align_size(tlv.length)).pack(*tlv)


def _tree_to_bin(
    tree: _TreeNode,
) -> bytes:
    """Encode a whole tree to binary TLVs"""
    # Go through the tree and encode to TLV->bin each branch/leaf
    if tree.leaf_value is not None:
        tlv = _Tlv(tree.osef_type, len(tree.leaf_value), tree.leaf_value)
    else:
        out = bytearray()
        for child in tree.children:
            out += _tree_to_bin(child)
        tlv = _Tlv(tree.osef_type, len(out), out)
    return _encode_tlv(tlv)


def _pack_to_tree(osef_type: Union[_types.OsefTypes, str], value: Any) -> _TreeNode:
    """Parse an item and generate a TreeNode, using OSEF types"""
    # deactivate unpacking to avoid unpacking binary values
    if isinstance(value, (_LazyDict, _LazyList)):
        value.unpacking = False
    if isinstance(osef_type, str):
        osef_type = _types.get_type_by_key(osef_type)
    type_info = _types.get_type_info_by_id(osef_type.value)

    # For leaves or unknown, return value
    if isinstance(type_info.node_info, _packers.PackerBase):
        return _TreeNode(osef_type.value, None, _pack_value(value, type_info.node_info))

    # For non-leaves, add each child to a list
    children = []
    if type_info.node_info.node_type == list:
        for child in value:
            child_k, child_v = list(child.items())[0]
            children.append(_pack_to_tree(child_k, child_v))
    elif type_info.node_info.node_type == dict:
        for child_k, child_v in value.items():
            children.append(_pack_to_tree(child_k, child_v))
    else:
        raise ValueError("Unsupported internal node type.")

    # reactivate unpacking
    if isinstance(value, (_LazyDict, _LazyList)):
        value.unpacking = True

    return _TreeNode(osef_type.value, children, None)


def _pack_value(value: Any, node_packer: _packers.PackerBase) -> bytes:
    """Pack a leaf value to a python object (type depends on type of leaf).

    :param value: object to be packed.
    :param node_packer: Packer for packing and conversion to python object.
    :return: bytes
    """
    try:
        if isinstance(value, BinaryLeaf):
            return value.leaf_value
        if node_packer.pack is not None:
            return node_packer.pack(value)
        # unknown parser
        return value
    except Exception:  # pylint: disable=broad-except
        osef_logger.error(
            f"Exception occurred while packing value for {node_packer}:\n"
            f"Details: {traceback.format_exc()}"
        )
        return value



"""Functions to read and parse osef files/streams."""
from typing import Tuple, Any

from osef.spec import _types
from osef.spec.constants import _TreeNode
from osef.parsing._parser_common import unpack_value
from osef.spec import _packers


def parse_to_dict(frame_tree: _TreeNode) -> dict:
    """Parse a whole frame tree to a python dictionary. All values of the tree will be unpacked.

    :param frame_tree: raw tree of a tlv frame.
    :return: dictionary with all values in osef frame.
    """
    type_name, subtree = _parse_raw_to_tuple(frame_tree)
    return {type_name: subtree}


def _parse_raw_to_tuple(raw_tree: _TreeNode) -> Tuple[str, Any]:
    """Parse a raw TLV tree, using OSEF types"""
    osef_type, children, leaf_value = raw_tree

    # Get leaf type info
    type_info = _types.get_type_info_by_id(osef_type)

    # For leaves or unknown, return value
    if isinstance(type_info.node_info, _packers.PackerBase):
        return (
            type_info.name,
            unpack_value(leaf_value, type_info.node_info, type_info.name),
        )

    # For non-leaves, add each child to a dictionary
    tree = [] if type_info.node_info.node_type == list else {}

    for child in children:
        child_name, child_tree = _parse_raw_to_tuple(child)

        if type_info.node_info.node_type == dict:
            tree[child_name] = child_tree
        elif type_info.node_info.node_type == list:
            tree.append({child_name: child_tree})
        else:
            raise ValueError("Unsupported internal node type.")

    return type_info.name, tree



"""Custom exceptions for OSEF related errors"""


class FieldError(KeyError):
    """Exception when a field is missing in an OSEF frame."""

    def __init__(self, field_key: str):
        """Raise an exception when a field is missing in an OSEF frame.

        :param field_key: Field key that raised the exception.
        """
        super().__init__(f"OSEF frame does not contain any '{field_key}' field.")


class ObjectIdError(ValueError):
    """Exception when an object ID is not in an OSEF frame."""

    def __init__(self, object_id: int):
        """Raise an exception when an object ID is missing in an OSEF frame.

        :param object_id: Object ID that raised the exception.
        """
        super().__init__(
            f"OSEF frame does not contain any object with ID '{object_id}'."
        )



"""Module to edit the data of an OSEF frame."""

# Standard imports
from typing import Dict, Optional, Union

# Third-party imports
import numpy as np
from numpy import typing as npt

# Local imports
from osef._logger import osef_logger
from osef.spec import constants
from osef.frame_helper.exceptions import FieldError, ObjectIdError
from osef.spec.osef_types import OsefKeys

# Constants
SHAPE_2D = 2
SHAPE_3D = 3


def set_timestamp(frame_dict: Dict, timestamp: float) -> None:
    """Set the timestamp of an OSEF frame.

    :param frame_dict: OSEF frame dictionary to edit.
    :param timestamp: timestamp value to set (in seconds)
    """
    if not isinstance(timestamp, float):
        raise ValueError("Invalid type for setting timestamp")

    frame_dict[OsefKeys.TIMESTAMPED_DATA][OsefKeys.TIMESTAMP_MICROSECOND] = timestamp


def _get_scan_frame_dict(frame_dict: dict) -> Optional[Dict]:
    """Get the scan frame sub-dictionary of an OSEF frame.

    :param frame_dict: OSEF frame dictionary to get.
    :return: tracked objects sub-dictionary, None if the field is absent.
    """
    return frame_dict.get(OsefKeys.TIMESTAMPED_DATA, {}).get(OsefKeys.SCAN_FRAME)


def _get_aug_cloud_dict(frame_dict: dict) -> Optional[Dict]:
    """Get the AugmentedCloud sub-dictionary of an OSEF frame.

    :param frame_dict: OSEF frame dictionary.
    :return: augmented cloud sub-dictionary, None if the field is absent.
    """
    return (
        frame_dict.get(OsefKeys.TIMESTAMPED_DATA, {})
        .get(OsefKeys.SCAN_FRAME, {})
        .get(OsefKeys.AUGMENTED_CLOUD)
    )


def set_augmented_cloud_field(
    frame_dict: Dict, field_key: str, value: Union[int, np.ndarray]
) -> None:
    """
    Set the value of a field of the augmented cloud.

    This function will overwrite the previous value if the field already exists,
    else it will create the field.
    Warning:
    - Number of points cannot be set manually because it must reflect the content
    of the other augmented cloud fields.
    - Cartesian coordinates field must be an array with a (N, 3) shape

    :param frame_dict: OSEF frame dictionary to edit.
    :param field_key: Key of the field to set in the augmented cloud.
    :param value: value of the field to set.
    """
    if (aug_cloud_dict := _get_aug_cloud_dict(frame_dict)) is None:
        raise FieldError(OsefKeys.AUGMENTED_CLOUD.value)

    if field_key == OsefKeys.NUMBER_OF_POINTS:
        raise ValueError(f"Field {OsefKeys.NUMBER_OF_POINTS} cannot be set manually.")

    if isinstance(value, np.ndarray):
        expected_length = aug_cloud_dict[OsefKeys.NUMBER_OF_POINTS]
        if value.shape[0] != expected_length:
            raise ValueError(
                "'value' must have the same length as the number of points"
                + f" ({value.shape[0]} != {expected_length}"
            )
        if len(value.shape) == SHAPE_2D and value.shape[1] != SHAPE_3D:
            raise ValueError(
                f"Field {field_key} has not a valid shape."
                + f" ({value.shape} != {expected_length, SHAPE_3D})"
            )

    aug_cloud_dict[field_key] = value


def filter_cloud(frame_dict: Dict, condition: npt.NDArray[np.bool_]) -> None:
    """
    Filter the augmented cloud by keeping points on condition.

    This will filter all the fields contained in the augmented cloud and update the
    number of points automatically.

    :param frame_dict: OSEF frame dictionary to edit.
    :param condition: Array of boolean flags, if True the cooresponding point will be kept.
    """
    if (aug_cloud_dict := _get_aug_cloud_dict(frame_dict)) is None:
        raise FieldError(OsefKeys.AUGMENTED_CLOUD.value)

    expected_length = aug_cloud_dict[OsefKeys.NUMBER_OF_POINTS]
    if len(condition.shape) != 1 or condition.shape[0] != expected_length:
        raise ValueError(
            f"`condition` has not the right shape ({condition.shape} != ({expected_length},))"
        )

    aug_cloud_dict[OsefKeys.NUMBER_OF_POINTS] = np.count_nonzero(condition)
    for field_key, field_data in aug_cloud_dict.items():
        if isinstance(field_data, np.ndarray):
            # In the raw frame dict, the 'cartesian_coordinates' field has a (N, 3)
            # that can be filtered the same way
            set_augmented_cloud_field(frame_dict, field_key, field_data[condition])


def _get_tracked_objects_dict(frame_dict: dict) -> Dict:
    """Get the AugmentedCloud sub-dictionary of an OSEF frame.

    :param frame_dict: OSEF frame dictionary to get.
    :return: tracked objects sub-dictionary.
    :raises FieldError: if tracked_objects field is not present in the frame
    """
    try:
        return frame_dict.get(OsefKeys.TIMESTAMPED_DATA, {}).get(
            OsefKeys.SCAN_FRAME, {}
        )[OsefKeys.TRACKED_OBJECTS]
    except KeyError as exc:
        raise FieldError(OsefKeys.TRACKED_OBJECTS) from exc


def set_tracked_objects_field(
    frame_dict: Dict, field_key: str, value: np.ndarray
) -> None:
    """Set the value of a field of the tracked objects.

    Warning: the 'number_of_objects' field cannot be set manually, it must correspond
    to the length of the other fields.

    :param frame_dict: OSEF frame dictionary to edit.
    :param field_key: Key of the field to set in the tracked objects.
    :param value: value of the field to set.
    """
    tracked_object_dict = _get_tracked_objects_dict(frame_dict)

    if field_key == OsefKeys.NUMBER_OF_OBJECTS:
        raise ValueError(f"Field {OsefKeys.NUMBER_OF_OBJECTS} cannot be set manually.")

    expected_length = tracked_object_dict[OsefKeys.NUMBER_OF_OBJECTS]
    if value.shape[0] != expected_length:
        raise ValueError(
            "'value' must have the same length as the number of objects"
            + f" ({value.shape[0]} != {expected_length})"
        )

    tracked_object_dict[field_key] = value


def remove_object(frame_dict: Dict, object_id: int) -> None:
    """Remove a tracked object from an OSEF frame.

    :param frame_dict: OSEF frame dictionary to edit.
    :param object_id: ID of the object to remove from the tracked object list.
    """
    tracked_object_dict = _get_tracked_objects_dict(frame_dict)

    # Check if object is in the frame, and get its index.
    # Warning: If the frame contains duplicated object IDs, all of them will be removed
    object_indices = np.where(
        tracked_object_dict[OsefKeys.OBJECT_ID_32_BITS] == object_id
    )[0]
    if object_indices.shape[0] == 0:
        raise ObjectIdError(object_id)

    # Update the number of objects.
    tracked_object_dict[OsefKeys.NUMBER_OF_OBJECTS] -= object_indices.shape[0]

    # Remove zone bindings before removing the ID.
    if OsefKeys.ZONES_OBJECTS_BINDING_32_BITS in _get_scan_frame_dict(frame_dict):
        remove_object_binding(frame_dict, object_id)

    # Remove all data related to the object.
    for tracked_objects_key, tracked_object_field in tracked_object_dict.items():
        # Warning: All fields of the tracked objects will be filtered
        if isinstance(tracked_object_field, np.ndarray):
            new_field_value = np.delete(
                tracked_object_dict[tracked_objects_key], object_indices, axis=0
            )
            set_tracked_objects_field(frame_dict, tracked_objects_key, new_field_value)

    # Set the object ID to 0 in the augmented cloud
    if (
        aug_cloud_dict := _get_aug_cloud_dict(frame_dict)
    ) is not None and OsefKeys.OBJECT_ID_32_BITS in aug_cloud_dict:
        object_mask = aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS] == object_id

        # Check if the augmented_cloud field is writeable, else make a copy
        new_object_ids = (
            aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS]
            if aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS].flags.writeable
            else np.copy(aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS])
        )
        new_object_ids[object_mask] = 0
        set_augmented_cloud_field(
            frame_dict, OsefKeys.OBJECT_ID_32_BITS, new_object_ids
        )


def remove_objects(frame_dict: dict, mask: npt.NDArray[np.bool_]) -> None:
    """Remove tracked objects from an OSEF frame.

    :param frame_dict: OSEF frame dictionary to edit.
    :param mask: mask of object to keep (False: remove object, True: keep object)
    """
    tracked_object_dict = _get_tracked_objects_dict(frame_dict)

    if mask.size != tracked_object_dict[OsefKeys.NUMBER_OF_OBJECTS]:
        raise ValueError("Given mask must be of size number_of_objects")

    # Update the number of objects.
    tracked_object_dict[OsefKeys.NUMBER_OF_OBJECTS] = np.count_nonzero(mask)

    # Remove zone bindings before removing the ID.
    if OsefKeys.ZONES_OBJECTS_BINDING_32_BITS in _get_scan_frame_dict(frame_dict):
        for object_id in tracked_object_dict[OsefKeys.OBJECT_ID_32_BITS][mask]:
            remove_object_binding(frame_dict, object_id)

    # Set the object ID to 0 in the augmented cloud
    if (
        aug_cloud_dict := _get_aug_cloud_dict(frame_dict)
    ) is not None and OsefKeys.OBJECT_ID_32_BITS in aug_cloud_dict:
        object_mask = np.isin(
            aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS],
            np.invert(tracked_object_dict[OsefKeys.OBJECT_ID_32_BITS][mask]),
        )

        # Check if the augmented_cloud field is writeable, else make a copy
        new_object_ids = (
            aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS]
            if aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS].flags.writeable
            else np.copy(aug_cloud_dict[OsefKeys.OBJECT_ID_32_BITS])
        )
        new_object_ids[object_mask] = 0
        set_augmented_cloud_field(
            frame_dict, OsefKeys.OBJECT_ID_32_BITS, new_object_ids
        )

    # Remove all data related to the object.
    for tracked_objects_key, tracked_object_field in tracked_object_dict.items():
        # Warning: All fields of the tracked objects will be filtered
        if isinstance(tracked_object_field, np.ndarray):
            new_field_value = tracked_object_dict[tracked_objects_key][mask]
            set_tracked_objects_field(frame_dict, tracked_objects_key, new_field_value)


def remove_object_binding(frame_dict: Dict, object_id: int) -> None:
    """Remove zone-object bindings of a given object ID.

    :param frame_dict: OSEF frame dictionary to edit.
    :param object_id: ID of the object to remove from the zone binding field.
    """
    tracked_object_dict = _get_tracked_objects_dict(frame_dict)

    if object_id not in tracked_object_dict[OsefKeys.OBJECT_ID_32_BITS]:
        raise ObjectIdError(object_id)

    # Get all bindings of the object ID.
    scan_frame_dict = frame_dict[OsefKeys.TIMESTAMPED_DATA][OsefKeys.SCAN_FRAME]
    if OsefKeys.ZONES_OBJECTS_BINDING in scan_frame_dict:
        osef_logger.warning(
            f"'{OsefKeys.ZONES_OBJECTS_BINDING}' is deprecated, "
            + f"'field should be {OsefKeys.ZONES_OBJECTS_BINDING_32_BITS}' instead."
        )
        binding_key = OsefKeys.ZONES_OBJECTS_BINDING
    elif OsefKeys.ZONES_OBJECTS_BINDING_32_BITS in scan_frame_dict:
        binding_key = OsefKeys.ZONES_OBJECTS_BINDING_32_BITS
    else:
        raise FieldError(OsefKeys.ZONES_OBJECTS_BINDING_32_BITS)

    zone_bindings = scan_frame_dict.get(binding_key)
    remaining_binding_msk = (
        zone_bindings[constants.ExtraZoneBindingKeys.OBJECT_ID] != object_id
    )
    scan_frame_dict[binding_key] = zone_bindings[remaining_binding_msk]



"""
Object Oriented Programming classes and helpers
to convert OSEF type-based data into object-based dataholders.
"""

# Standard imports
from __future__ import annotations
from typing import Optional, NamedTuple

# Third-party imports
import numpy as np
from numpy import typing as npt

# Local imports
from osef.spec import osef_types


# Sign matrix to apply to bounding box dimension matrix to get the vertices
_BBOX_SIGNS = np.array(
    [
        [1, 1, -1, -1, 1, 1, -1, -1],
        [1, -1, -1, 1, 1, -1, -1, 1],
        [-1, -1, -1, -1, 1, 1, 1, 1],
    ]
)
_BBOX_FOOTPRINT_SIGNS = _BBOX_SIGNS[:2, :4]


class Pose(NamedTuple):
    """Class to handle a Pose from OSEF data."""

    translation: npt.NDArray[np.float32]
    rotation: npt.NDArray[np.float32]

    def __eq__(self, other: Pose) -> bool:
        """Equality operator."""
        return (np.array_equal(self.rotation, other.rotation)) and (
            np.array_equal(self.translation, other.translation)
        )

    def __hash__(self) -> int:
        """Hash method."""
        return hash((self.translation, self.rotation))

    @property
    def matrix(self) -> npt.NDArray[np.float32]:
        """Get a Matrix 4x4 with the rotation and translation."""
        pose_matrix = np.identity(4)
        pose_matrix[:3, :3] = self.rotation
        pose_matrix[:3, 3] = self.translation
        return pose_matrix


class GeographicPose(NamedTuple):
    """Class to handle a Geographic Pose from OSEF data."""

    latitude: np.float64
    longitude: np.float64
    heading: np.float32


class GeographicSpeed(NamedTuple):
    """Class to handle a Geographic Speed from OSEF data."""

    speed: np.float32
    heading: np.float32


class ObjectProperties(NamedTuple):
    """Class to handle the object properties."""

    oriented: bool
    is_seen: bool
    has_valid_slam_pose: bool
    is_static: bool


class ZoneBindings(NamedTuple):
    """Class to handle the zone bindings."""

    zone_index: int
    object_id: int

    def __repr__(self) -> str:
        """String representation of the Zone binding class."""
        return f"Binding [Zone {self.zone_index} - Object {self.object_id}]"


class ZoneDefinition(NamedTuple):
    """Class to handle zone definition."""

    name: str
    vertices: npt.NDArray[np.float32]
    vertical_limits: Optional[npt.NDArray[np.float32]] = None


class TrackedObjectData(NamedTuple):
    """Helper class for the data of one Tracked Object."""

    object_id: np.uint32
    class_id: np.int32
    speed_vector: npt.NDArray[np.float32]
    bbox_size: npt.NDArray[np.float32]
    translation: npt.NDArray[np.float32]
    rotation: npt.NDArray[np.float32]
    properties: Optional[ObjectProperties] = None
    geographic_pose: Optional[GeographicPose] = None
    geographic_speed: Optional[GeographicSpeed] = None
    classification_confidence: Optional[np.float32] = None

    @property
    def class_name(self) -> str:
        """Name of the object class."""
        return osef_types.ClassId(self.class_id).name

    @property
    def bbox_footprint_vertices(self) -> npt.NDArray[np.float32]:
        """
        Compute the coordinates of the projected bounding box on the ground plane.
        The output array has a (4, 2) shape.
        """
        bbox_local_coordinates = self.bbox_size[:2][:, None] * _BBOX_FOOTPRINT_SIGNS / 2
        return (
            np.dot(self.rotation[:2, :2], bbox_local_coordinates)
            + self.translation[:2][:, None]
        ).T

    @property
    def bbox_vertices(self) -> npt.NDArray[np.float32]:
        """
        Compute the coordinates of the bounding box vertices.
        The output array has a (8, 3) shape.
        """
        bbox_local_coordinates = self.bbox_size[:, None] * _BBOX_SIGNS / 2
        return (
            np.dot(self.rotation, bbox_local_coordinates) + self.translation[:, None]
        ).T



"""Scanning tools to speed up OSEF frame access"""

# Standard imports
import pathlib
from struct import Struct
from typing import Union, List
from urllib.parse import urlparse

# Project imports
from osef._logger import osef_logger
from osef.spec import constants
from osef.parsing._parser_common import _read_from_file
from osef.parsing import osef_stream


def find_frame_file_offsets(
    osef_path: Union[str, pathlib.Path], log_period: int = -1
) -> List[int]:
    """Walk through OSEF file and find start position of each frame (frame_offset).
    The position in the file is useful to easily jump to a frame.

    :param osef_path: path to OSEF file to scan
    :param log_period: will print a log of the number of frames read (-1 by default which does not print)
    :return: frame_offsets: list of file offsets of each OSEF frame
    """
    frame_offsets = [0]
    if urlparse(str(osef_path)).scheme == osef_stream.TCP_SCHEME:
        raise ValueError(
            f"{find_frame_file_offsets.__name__} is not compatible with TCP streams"
        )

    with osef_stream.OsefStreamFile(osef_path) as stream:
        osef_logger.info(
            "Start scanning OSEF file (find each frame positions in file)."
        )
        while True:
            try:
                # Read header
                struct = Struct(constants._STRUCT_FORMAT % 0)
                blob = _read_from_file(stream, struct.size)

                # Unpack Type and Length and jump to next frame
                read_tlv = constants._Tlv._make(struct.unpack_from(blob))
                stream._io_stream.seek(stream._io_stream.tell() + read_tlv.length)
            except EOFError:
                break
            frame_offsets.append(stream._io_stream.tell())

            # Log every COUNT_FREQUENCY_UPDATE frames
            if log_period >= 0 and not len(frame_offsets) % log_period:
                osef_logger.info(
                    f"OSEF file indexing progress : {len(frame_offsets)} frames"
                )

        frame_offsets = frame_offsets[:-1]  # remove last offset

        osef_logger.info(
            f"Finished scanning OSEF, file size: {len(frame_offsets)} frames"
        )
        return frame_offsets


def count_frames(osef_path: Union[str, pathlib.Path], log_period: int = -1):
    """Count the number of frames in an OSEF file.

    :param osef_path: path to OSEF file to scan
    :param log_period: will print a log of the number of frames read (-1 by default which does not print)
    :return: the number of OSEF frames in the file
    """
    return len(find_frame_file_offsets(osef_path, log_period))



"""Functions to read and parse osef files/streams."""
import struct
import traceback
from collections import deque
from itertools import islice
from struct import Struct
from typing import Any, Iterable, Optional, Tuple, Iterator

from osef.spec import _types
from osef._logger import osef_logger
from osef.spec.constants import _Tlv, _TreeNode, _STRUCT_FORMAT
from osef.parsing.osef_stream import OsefStream, OsefStreamTcp
from osef.spec import _packers


class MalformedTlvException(Exception):
    """Exception raised for TLV structures that are malformed or incorrectly shaped."""


class OsefParsingException(Exception):
    """Exception raised when failing to parse next dataframe."""


def unpack_value(
    value: bytes, node_packer: _packers.PackerBase, type_name: str = ""
) -> Any:
    """Unpack a leaf value to a python object (type depends on type of leaf).

    :param value: binary value to be unpacked.
    :param node_packer: Packer for unpacking and conversion to python object.
    :param type_name: (optional) provide type name
     to provide better feedback if an exception occurs
    :return: python object
    """
    try:
        if node_packer.unpack is not None:
            return node_packer.unpack(value)
        # unknown parser
        return value

    except Exception as err:
        raise type(err)(f'occurred while unpacking "{type_name}".') from err


def iter_file(osef_stream: OsefStream, auto_reconnect: bool = True) -> Iterator[_Tlv]:
    """Iterator function to iterate over each frame in osef file.

    :param osef_stream: opened binary file containing tlv frames.
    :param auto_reconnect: enable reconnection for tcp connections.
    :return frame_tlv: next tlv frame of the osef file.
    """
    while True:
        try:
            frame_tlv = read_next_tlv(osef_stream)
        except EOFError:
            if auto_reconnect and isinstance(osef_stream, OsefStreamTcp):
                osef_logger.warning("Connection lost: reopening socket")
                osef_stream.open_socket(auto_reconnect)
                continue
            break
        except Exception:  # pylint: disable=broad-except
            osef_logger.error(
                f"Error: cannot read next Tlv from file (malformed Tlv?).\n"
                f"Details: {traceback.format_exc()}\n"
            )
            break

        yield frame_tlv


def get_tlv_iterator(
    opened_file: OsefStream,
    first: Optional[int] = None,
    last: Optional[int] = None,
    auto_reconnect: Optional[bool] = True,
) -> Iterable[_Tlv]:
    """Get an iterator to iterate over each tlv frame in osef file.

    :param opened_file: opened binary file containing tlv frames.
    :param first: iterate only on N first frames of file.
    :param last: iterate only on M last frames of file.
    Can be used with first to get the range (N-M) -> N
    :param auto_reconnect: enable reconnection for tcp connections.
    :return: tlv frame iterator
    """
    if first is None and last is None:
        return enumerate(iter_file(opened_file, auto_reconnect))
    return deque(islice(enumerate(iter_file(opened_file, auto_reconnect)), first), last)


def build_tree(tlv: _Tlv) -> _TreeNode:
    """Recursive function to get a tree from a raw Tlv frame

    :param tlv: raw tlv frame read from file.
    :raises  MalformedTlvException when _parse_tlv_from_blob raises a struct.error
    :return: tree representation of the tlv frame
    """
    # If we know this type is an internal node (not a leaf)\
    if tlv.osef_type in _types.outsight_types and isinstance(
        _types.outsight_types[tlv.osef_type].node_info, _types.InternalNodeInfo
    ):
        read = 0
        children = []
        while read < tlv.length:
            try:
                sub_tlv, sub_size = _parse_tlv_from_blob(tlv.value, read)
            except struct.error as exception:
                raise MalformedTlvException(
                    "Malformed Tlv, unable to unpack values."
                ) from exception

            sub_tree = build_tree(sub_tlv)
            children.append(sub_tree)
            read += sub_size
        return _TreeNode(tlv.osef_type, children, None)
    return _TreeNode(tlv.osef_type, None, tlv.value)


def read_next_tlv(osef_stream: OsefStream) -> Optional[_Tlv]:
    """Read the next TLV from a binary stream (file or socket)"""
    # Read header
    structure = Struct(_STRUCT_FORMAT % 0)
    blob = _read_from_file(osef_stream, structure.size)
    # Parse Type and Length
    read_tlv = _Tlv._make(structure.unpack_from(blob))

    # Now that we know its length we can read the Value
    structure = Struct(_STRUCT_FORMAT % read_tlv.length)
    blob += _read_from_file(osef_stream, structure.size - len(blob))
    read_tlv = _Tlv._make(structure.unpack_from(blob))

    return read_tlv


def _align_size(size: int) -> int:
    """Returned aligned size from tlv size"""
    alignment_size = 4
    offset = size % alignment_size
    return size if not offset else size + alignment_size - offset


def _read_from_file(osef_stream: OsefStream, byte_number: int) -> bytes:
    """Read given number of bytes from readable stream"""
    blob = bytearray()
    while len(blob) < byte_number:
        blob_inc = osef_stream.read(byte_number - len(blob))
        # End of file
        if blob_inc is None or len(blob_inc) == 0:
            raise EOFError
        blob += blob_inc
    return blob


def _parse_tlv_from_blob(blob: bytes, offset=0) -> Tuple[_Tlv, int]:
    """Parse a TLV from a binary blob"""
    # Unpack a first time to get Type and Length
    structure = Struct(_STRUCT_FORMAT % 0)
    read_tlv = _Tlv._make(structure.unpack_from(blob, offset))

    # Then unpack the whole tlv
    structure = Struct(_STRUCT_FORMAT % read_tlv.length)
    read_tlv = _Tlv._make(structure.unpack_from(blob, offset))
    return read_tlv, _align_size(structure.size)



"""

Library containing utilities to read and parse a stream, live or recorded, retrieved from
**Shift**.

"""
# Import version
from osef._version import __version__

# Shortcuts for public functions/classes
from osef.frame_helper import osef_frame
from osef.parsing.parser import parse
from osef.packing.packer import pack



"""Constants used throughout project and that can be used by user"""

# Standard imports
from __future__ import annotations
from typing import List, NamedTuple, Optional

# Project imports
from osef.spec import osef_types

# Structure Format definition (see https://docs.python.org/3/library/struct.html#format-strings):
# Meant to be used as: _STRUCT_FORMAT % length

_STRUCT_FORMAT = "<"  # little endian
_STRUCT_FORMAT += "L"  # unsigned long        (field 'T' ie. 'Type')
_STRUCT_FORMAT += "L"  # unsigned long        (field 'L' ie. 'Length')
_STRUCT_FORMAT += "%ds"  # buffer of fixed size (field 'V' ie. 'Value')


class _Tlv(NamedTuple):
    """Type-Length-Value (TLV) definition.

    :param osef_type: Integer value for defining the OSEF type.
    :param length: Length of the data.
    :param value: Actual data.
    """

    osef_type: osef_types.OsefTypes
    length: int
    value: bytes


class _TreeNode(NamedTuple):
    """Tree representation of the TLV data.

    :param osef_type: Type of the data.
    :param children: Optional tree of data's child.
    :param leaf_value: Data's value if no child.
    """

    osef_type: osef_types.OsefTypes
    children: Optional[List[_TreeNode]]
    leaf_value: Optional[bytes]


class ExtraPoseKeys:
    """Class to define extra OSEF keys, not defined in the spec,
    relative to the Pose.
    """

    ROTATION = "rotation"
    TRANSLATION = "translation"


class ExtraZoneBindingKeys:
    """Class to define extra OSEF keys, not defined in the spec,
    relative to the zone bindings.
    """

    ZONE_INDEX = "zone_idx"
    OBJECT_ID = "object_id"


class ExtraTimestampKeys:
    """Class to define extra OSEF keys, not defined in the spec,
    relative to the timestamp.
    """

    TIMESTAMP = "timestamp"
    UNIX_S = "unix_s"
    REMAINING_US = "remaining_us"


class ExtraImuKeys:
    """Class to define extra OSEF keys, not defined in the spec,
    relative to the IMU.
    """

    ACCELERATION = "acceleration"
    ANGULAR_VELOCITY = "angular_velocity"


class ExtraBackgroundKeys:
    """Class to define extra OSEF keys, not defined in the spec,
    relative to the background.
    """

    DELETED = "background_deleted"
    WIDTH = "width"
    HEIGHT = "height"
    FIRST_AZIMUTH = "first_azimuth"
    AZIMUTH_STEP = "azimuth_step"
    FIRST_INDEX = "firsr_index"
    CELLS_NUMBER = "cells_number"


class ExtraOsefKeys:
    """Keys for specifying some OSEF data/fields, which
    are not defined in the spec.
    """

    _TO_DROP = "__todrop"
    IMAGE_WIDTH = "image_width"
    IMAGE_HEIGHT = "image_height"
    AZIMUTH_BEGIN_DEG = "azimuth_begin_deg"
    AZIMUTH_END_DEG = "azimuth_end_deg"


class ExtraCoordinatesKeys:
    """Class to define extra OSEF keys, not defined in the spec,
    relative to the several coordinates data.
    """

    X = "x"
    Y = "y"
    Z = "z"
    X_MIN = "x_min"
    X_MAX = "x_max"
    Y_MIN = "y_min"
    Y_MAX = "y_max"
    AZIMUTH = "azimuth"
    ELEVATION = "elevation"
    DISTANCE = "distance"
    LATITUDE = "lat"
    LONGITUDE = "long"
    HEADING = "heading"
    SPEED = "speed"


class ObjectPropertiesIndexes:
    """Class to define the object properties correspondance
    between index and property.
    """

    ORIENTED = 0
    IS_SEEN = 1
    HAS_VALID_SLAM_POSE = 2
    IS_STATIC = 3



"""OSEF private types definition."""
__version__ = "1.17.0"

# Standard imports
from enum import Enum, IntEnum

# !!! Warning !!!
# !!! This file has been auto generated, do not attempt to edit it !!!
# Important: all words use little-endian representation, unless specified otherwise


class OsefTypes(IntEnum):
    """OSEF types."""

    AUGMENTED_CLOUD = 1
    NUMBER_OF_POINTS = 2
    SPHERICAL_COORDINATES = 3
    REFLECTIVITIES = 4
    _BACKGROUND_FLAGS = 5
    CARTESIAN_COORDINATES = 6
    _BGR_COLORS = 7
    _OBJECT_DETECTION_FRAME = 8
    _IMAGE_DIMENSION = 9
    NUMBER_OF_OBJECTS = 10
    _CLOUD_FRAME = 11
    TIMESTAMP_MICROSECOND = 12
    _AZIMUTHS_COLUMN = 13
    NUMBER_OF_LAYERS = 14
    _CLOUD_PROCESSING = 15
    _RANGE_AZIMUTH = 16
    _BOUNDING_BOXES_ARRAY = 17
    CLASS_ID_ARRAY = 18
    _CONFIDENCE_ARRAY = 19
    TIMESTAMPED_DATA = 20
    _PERCEPT = 21
    _BGR_IMAGE_FRAME = 23
    POSE = 24
    SCAN_FRAME = 25
    TRACKED_OBJECTS = 26
    BBOX_SIZES = 27
    SPEED_VECTORS = 28
    POSE_ARRAY = 29
    OBJECT_ID = 30
    CARTESIAN_COORDINATES_4F = 31
    SPHERICAL_COORDINATES_4F = 32
    ZONES_DEF = 33
    ZONE = 34
    ZONE_VERTICES = 35
    ZONE_NAME = 36
    _ZONE_UUID = 37
    ZONES_OBJECTS_BINDING = 38
    OBJECT_PROPERTIES = 39
    _IMU_PACKET = 40
    _TIMESTAMP_LIDAR_VELODYNE = 41
    POSE_RELATIVE = 42
    _GRAVITY = 43
    EGO_MOTION = 44
    _PREDICTED_POSITION = 45
    GEOGRAPHIC_POSE = 46
    OBJECT_ID_32_BITS = 47
    ZONES_OBJECTS_BINDING_32_BITS = 48
    _BACKGROUND_BITS = 49
    _GROUND_PLANE_BITS = 50
    _AZIMUTHS = 51
    _ELEVATIONS = 52
    _DISTANCES = 53
    _LIDAR_MODEL = 54
    SLAM_POSE_ARRAY = 55
    ZONE_VERTICAL_LIMITS = 56
    GEOGRAPHIC_POSE_PRECISE = 57
    _ROAD_MARKINGS_BITS = 58
    SMOOTHED_POSE = 59
    _INTERACTIVE_REQUEST = 60
    _INTERACTIVE_RESPONSE = 61
    _INTERACTIVE_REQUEST_ID = 62
    _INTERACTIVE_REQUEST_BACKGROUND_HEADER = 63
    _INTERACTIVE_RESPONSE_BACKGROUND_HEADER = 64
    _INTERACTIVE_REQUEST_BACKGROUND_DATA = 65
    _INTERACTIVE_RESPONSE_BACKGROUND_DATA = 66
    _INTERACTIVE_REQUEST_PINGPONG = 67
    _INTERACTIVE_RESPONSE_PINGPONG = 68
    _HEIGHT_MAP = 69
    _HEIGHT_MAP_POINTS = 70
    DIVERGENCE_INDICATOR = 71
    _CARLA_TAG_ARRAY = 72
    _BACKGROUND_SCENE_PARAMS = 73
    _BACKGROUND_SCENE_PARAMS_GENERAL = 74
    _BACKGROUND_SCENE_PARAMS_ELEVATIONS = 75
    _BACKGROUND_SCENE_FRAGMENT = 76
    _BACKGROUND_SCENE_FRAGMENT_INFO = 77
    _BACKGROUND_SCENE_FRAGMENT_DISTANCES = 78
    GEOGRAPHIC_POSE_ARRAY = 79
    GEOGRAPHIC_SPEED = 80
    GEOGRAPHIC_SPEED_ARRAY = 81
    _INSTANTANEOUS_TRANSLATION_SPEED = 82
    _INSTANTANEOUS_ROTATION_SPEED = 83
    _FILTERED_TRANSLATION_SPEED = 84
    _FILTERED_ROTATION_SPEED = 85
    REFERENCE_MAP_BITS = 86
    _CARTESIAN_COVARIANCE = 87
    _CYLINDRICAL_COVARIANCE = 88
    COORDINATES_REFERENCE_SYSTEM = 89
    START_TIMESTAMP = 90
    END_TIMESTAMP = 91
    TIME_DOMAIN = 92
    TIME_INTERVAL = 93
    CLASSIFICATION_CONFIDENCE = 94
    _CLASS_PROBABILITIES_ARRAY = 95
    _CLASSIFIER_DEF = 96
    _ENABLED_CLASS_IDS_ARRAY = 97
    _REFLECTION_FEATURES_ARRAY = 98
    _REFLECTION_BITS = 99


class OsefKeys(str, Enum):
    """OSEF type keys."""

    def __repr__(self) -> str:
        """Override repr method to display OsefKeys directly as strings."""
        return f"'{self.value}'"

    AUGMENTED_CLOUD = "augmented_cloud"
    """
    Augmented Cloud
    **Output when**: Augmented Cloud is enabled.
    **Purpose**: An augmented cloud represents a cloud of points. Each point can have attributes, all of which are
    optional, which describe its position, physical property, movement, etc...
    Contains following sub-TLVs:
    - exactly one:     Number of Points               (NUMBER_OF_POINTS)
    - exactly one:     Number of Layers               (NUMBER_OF_LAYERS)
    - none or one:     Spherical Coordinates          (SPHERICAL_COORDINATES) [DEPRECATED USAGE]
    - none or one:     Reflectivities                 (REFLECTIVITIES)
    - none or one:     Background Flags               (BACKGROUND_FLAGS) [DEPRECATED USAGE]
    - none or one:     Cartesian Coordinates          (CARTESIAN_COORDINATES)
    - none or one:     BGR Colors                     (BGR_COLORS)
    - none or one:     Azimuths Column                (AZIMUTHS_COLUMN)
    - none or one:     Cloud Processing               (CLOUD_PROCESSING)
    - none or one:     Percept                        (PERCEPT) [DEPRECATED USAGE]
    - none or one:     Object ID 32 Bits              (OBJECT_ID_32_BITS)
    - none or one:     Cartesian Coordinates 4F       (CARTESIAN_COORDINATES_4F) [DEPRECATED USAGE]
    - none or one:     Background Bits                (BACKGROUND_BITS)
    - none or one:     Ground Plane Bits              (GROUND_PLANE_BITS)
    - none or one:     Azimuths                       (AZIMUTHS)
    - none or one:     Elevations                     (ELEVATIONS)
    - none or one:     Distances                      (DISTANCES)
    - none or one:     Road Markings Bits             (ROAD_MARKINGS_BITS) [DEPRECATED USAGE]
    - none or one:     Carla Tag Array                (CARLA_TAG_ARRAY)
    - none or one:     Reference Map Bits             (REFERENCE_MAP_BITS)
    - none or one:     Coordinates Reference System   (COORDINATES_REFERENCE_SYSTEM)
    - none or one:     Reflection Bits                (REFLECTION_BITS)
    """

    NUMBER_OF_POINTS = "number_of_points"
    """
    Number of Points
    **Output when**: Augmented Cloud is enabled.
    **Format**: Contains 32 bits unsigned int value.
    **Purpose**: Represents the number of points in the point cloud.
    """

    SPHERICAL_COORDINATES = "spherical_coordinates"
    """
    Spherical Coordinates [DEPRECATED USAGE]
    Contains list of spherical single-precision float coordinates, three per point:
    * azimuth : degrees, range [ 0.0 .. 360.0 ]
    * elevation : degrees, range [ -90.0 .. +90.0 ]
    * distance : meters, range [ 0.0 ..  +inf ]
    """

    REFLECTIVITIES = "reflectivities"
    """
    Reflectivities
    **Output when**: Augmented Cloud is enabled.
    **Format**: List of unsigned 8-bits integers.
    **Purpose**: Contains list of reflectivity values.
    Diffuse reflectors report values from 0 to 100 for reflectivities from 0% to 100%.
    Retroreflectors report values from 101 to 255, where 255 represents an ideal reflection.
    """

    _BACKGROUND_FLAGS = "background_flags"
    """
    Background Flags [DEPRECATED USAGE]
    **Instead, use**: "Background bits" (Type 49)
    Contains a list of boolean values encoded on an unsigned 8-bits integer:
    * 0 means the point is not part of the background
    * all other values mean the point is part of the background
    The background is defined as the objects which were frequently detected in the past.
    On the contrary the foreground is composed of objects that were not seen before (i.e. they appeared recently and/or
    they are moving).
    """

    CARTESIAN_COORDINATES = "cartesian_coordinates"
    """
    Cartesian Coordinates
    **Output when**: Augmented Cloud is enabled.
    Contains list of Cartesian single-precision float coordinates, three per point:
    * x : meters, range [ -inf .. +inf ]
    * y : meters, range [ -inf .. +inf ]
    * z : meters, range [ -inf .. +inf ]
    **Coordinate Frame**:
    - Passthrough & Mobile: Relative to LiDAR Coordinate Frame.
    - Static: in Absolute/World Frame, which is:
        - Single-LiDAR: On-Ground LiDAR Coordinate Frame.
        - Multi-LiDARs: Map Coordinate Frame.
    """

    _BGR_COLORS = "bgr_colors"
    """
    BGR Colors
    Contain list of BGR colors, encoded as three consecutive unsigned 8-bits integers:
    * blue  : range [ 0 .. 255 ]
    * green : range [ 0 .. 255 ]
    * red   : range [ 0 .. 255 ]
    """

    _OBJECT_DETECTION_FRAME = "object_detection_frame"
    """
    Object Detection Frame
    Group all classification results relative to an image.
    Contains following sub-TLVs:
    - exactly one:     Number of Objects              (NUMBER_OF_OBJECTS)
    - none or one:     Timestamp Microsecond          (TIMESTAMP_MICROSECOND)
    - exactly one:     Bounding Boxes Array           (BOUNDING_BOXES_ARRAY)
    - exactly one:     Class ID Array                 (CLASS_ID_ARRAY)
    - none or one:     Confidence Array               (CONFIDENCE_ARRAY)
    """

    _IMAGE_DIMENSION = "image_dimension"
    """
    Image Dimension
    Contain following concatenated data:
    * image_width   : in pixel, unsigned 32-bits integer, range [ 0 .. 2^32 [
    * image_height  : in pixel, unsigned 32-bits integer, range [ 0 .. 2^32 [
    """

    NUMBER_OF_OBJECTS = "number_of_objects"
    """
    Number of Objects
    **Output when**: Tracking is enabled.
    **Purpose**: Number of tracked objects.
    **Format**: Unsigned 32-bits integer, range [ 0 .. 2^32 [
    """

    _CLOUD_FRAME = "cloud_frame"
    """
    Cloud Frame
    Contains following sub-TLVs:
    - exactly one:     Augmented Cloud                (AUGMENTED_CLOUD)
    - exactly one:     Range Azimuth                  (RANGE_AZIMUTH)
    - none or one:     Timestamp Lidar Velodyne       (TIMESTAMP_LIDAR_VELODYNE)
    - none or one:     LiDAR Model                    (LIDAR_MODEL)
    """

    TIMESTAMP_MICROSECOND = "timestamp_microsecond"
    """
    Timestamp Microsecond
    **Output when**: Always.
    **Purpose**: Describes a unix timestamp with microsecond precision, same as struct timeval of UNIX <sys/time.h>.
    **Format**: Contains concatenation of:
    * UNIX time in seconds,   unsigned 32-bits integer, range [ 0 .. 2^32 [
    * remaining microseconds, unsigned 32-bits integer, range [ 0 .. 1000000 [
    """

    _AZIMUTHS_COLUMN = "azimuths_column"
    """
    Azimuths Column
    Contains list of azimuth values in degrees expressed as single-precision floats.
    This azimuth value is computed before corrections due to:
      * The time between each laser firing
      * The correction when lasers of the same firing sequence are not aligned
    The range is [ 0.0 .. 360.0 [.
    """

    NUMBER_OF_LAYERS = "number_of_layers"
    """
    Number of Layers
    **Output when**: Augmented Cloud is enabled.
    **Format**: Contains 32 bits unsigned int value.
    **Purpose**: Represents the number of layers of the point cloud.
    0 indicates that this cloud does not have a 2D structure.
    """

    _CLOUD_PROCESSING = "cloud_processing"
    """
    Cloud Processing
    Contains a 64 bits bitfield, representing the processing that have been applied.
    If no bit are set this means that no processing have been done on this cloud.
    * Bit 0 : The background points have been removed
    """

    _RANGE_AZIMUTH = "range_azimuth"
    """
    Range Azimuth
    Range of azimuth values for the points contained in a given lidar packet.
    Contains exactly two 32 bits floats. The first one marks the beginning of the range, the second the end.
      The range is [ 0.0 .. 360.0 [.
    """

    _BOUNDING_BOXES_ARRAY = "bounding_boxes_array"
    """
    Bounding Boxes Array
    Contains list of bounding boxes dimensions, four per box:
    * x_min : percentage of the image size, 32-bits floating-point, range [ 0 .. 1 ]
    * y_min : percentage of the image size, 32-bits floating-point, range [ 0 .. 1 ]
    * x_max : percentage of the image size, 32-bits floating-point, range [ 0 .. 1 ]
    * y_max : percentage of the image size, 32-bits floating-point, range [ 0 .. 1 ]
    """

    CLASS_ID_ARRAY = "class_id_array"
    """
    Class ID Array
    **Output when**: Tracking is enabled.
    **Format**: List of signed 32 bits integers, using enum CLASS ID.
    **Purpose**: Contains list of class IDs. For further details, see:
    [Classification](https://docs.outsight.ai/software/understanding-osef-data/object-tracking-data#classification)
    """

    _CONFIDENCE_ARRAY = "confidence_array"
    """
    Confidence Array
    Contains list of confidences, one per element:
    32-bits floating-point, range [ 0.0 .. 1.0 ]
    """

    TIMESTAMPED_DATA = "timestamped_data"
    """
    Timestamped Data
    **Output when**: Always.
    Contains following sub-TLVs:
    - exactly one:     Timestamp Microsecond          (TIMESTAMP_MICROSECOND)
    - none or one:     Object Detection Frame         (OBJECT_DETECTION_FRAME)
    - none or one:     Cloud Frame                    (CLOUD_FRAME)
    - none or one:     Scan Frame                     (SCAN_FRAME)
    - zero to many:    Time interval                  (TIME_INTERVAL)
    """

    _PERCEPT = "percept"
    """
    Percept [DEPRECATED USAGE]
    Contains list of percept classes represented as unsigned 16-bits integer, using enum PERCEPT ID.
    """

    _BGR_IMAGE_FRAME = "bgr_image_frame"
    """
    BGR Image Frame
    Contains following sub-TLVs:
    - exactly one:     BGR Colors                     (BGR_COLORS)
    - exactly one:     Image Dimension                (IMAGE_DIMENSION)
    """

    POSE = "pose"
    """
    Pose
    **Output when**: Slam is enabled.
    **Purpose**: Pose of the LiDAR.
    **Format**: Concatenation of 12 floats of 32-bits each, in this order:
    - Tx, Ty and Tz, representing the translation vector T to get the position coordinates in meters
    - Rxx, Ryx, Rzx, Rxy, Ryy, Rzy, Rxz, Ryz and Rzz represent the rotation matrix R (given column-wise) defining the
    orientation
    As a consequence, considering a device in a pose defined by T and R, you can compute the coordinates Xabs in the
    absolute referential from the coordinates Xrel in device referential using the vectorial formula:
    ```
    Xabs = R * Xrel + T
    ```
    with:
    ```
        [Rxx Rxy Rxz]
    R = [Ryx Ryy Ryz]
        [Rzx Rzy Rzz]
        [Tx]
    T = [Ty]
        [Tz]
    ```
    **Coordinate Frame**:
    Absolute/World Frame, which is:
    - Ego-Motion: Initial pose of the LiDAR.
    - Relocalization: Map Coordinate Frame.
    """

    SCAN_FRAME = "scan_frame"
    """
    Scan Frame
    **Output when**: Always.
    Contains following sub-TLVs:
    - none or one:     Augmented Cloud                (AUGMENTED_CLOUD)
    - none or one:     Pose                           (POSE)
    - none or one:     Geographic Pose                (GEOGRAPHIC_POSE) [DEPRECATED USAGE]
    - none or one:     Geographic Pose Precise        (GEOGRAPHIC_POSE_PRECISE)
    - none or one:     Geographic Speed               (GEOGRAPHIC_SPEED)
    - none or one:     Ego Motion                     (EGO_MOTION)
    - none or one:     Tracked Objects                (TRACKED_OBJECTS)
    - none or one:     Zones Def                      (ZONES_DEF)
    - none or one:     Zones Objects Binding          (ZONES_OBJECTS_BINDING) [DEPRECATED USAGE]
    - none or one:     Zones Objects Binding 32 Bits  (ZONES_OBJECTS_BINDING_32_BITS)
    - none or one:     Height Map                     (HEIGHT_MAP)
    - none or one:     Classifier Def                 (CLASSIFIER_DEF)
    """

    TRACKED_OBJECTS = "tracked_objects"
    """
    Tracked Objects
    **Output when**: Tracking is enabled.
    **Purpose**: Properties of tracked objects, which includes centroid, bbox, speed...
    For further details, see: [Object Tracking
    Data](https://docs.outsight.ai/software/understanding-osef-data/object-tracking-data).
    Contains following sub-TLVs:
    - exactly one:     Number of Objects              (NUMBER_OF_OBJECTS)
    - exactly one:     Object ID 32 Bits              (OBJECT_ID_32_BITS)
    - exactly one:     Class ID Array                 (CLASS_ID_ARRAY)
    - exactly one:     Bbox Sizes                     (BBOX_SIZES)
    - exactly one:     Speed Vectors                  (SPEED_VECTORS)
    - exactly one:     Pose Array                     (POSE_ARRAY)
    - none or one:     Slam Pose Array                (SLAM_POSE_ARRAY)
    - exactly one:     Object Properties              (OBJECT_PROPERTIES)
    - none or one:     Geographic Pose Array          (GEOGRAPHIC_POSE_ARRAY)
    - none or one:     Geographic Speed Array         (GEOGRAPHIC_SPEED_ARRAY)
    - none or one:     Classification confidence      (CLASSIFICATION_CONFIDENCE)
    - none or one:     Class probabilities Array      (CLASS_PROBABILITIES_ARRAY)
    - none or one:     Reflection features array      (REFLECTION_FEATURES_ARRAY)
    """

    BBOX_SIZES = "bbox_sizes"
    """
    Bbox Sizes
    **Output when**: Tracking is enabled.
    **Purpose**: Define bounding boxes around each tracked object.
    The bounding box is extended/cropped in a way to always touch the ground along the z axis.
    A bounding box is defined by its 3 dimensions, on x, y, z axes, centered on Pose Array.
    **Format**: Array of bounding box sizes, three single-precision floats per object:
    * x : meters, range [ -inf .. +inf ]
    * y : meters, range [ -inf .. +inf ]
    * z : meters, range [ -inf .. +inf ]
    """

    SPEED_VECTORS = "speed_vectors"
    """
    Speed Vectors
    **Output when**: Tracking is enabled.
    **Purpose**: Speed vectors, one for each tracked object.
    Speed vectors are defined on x, y, z axes, centered on Pose Array.
    **Format**: Array of speed vectors, three single-precision floats per object:
    * x : meters per second, range [ -inf .. +inf ]
    * y : meters per second, range [ -inf .. +inf ]
    * z : meters per second, range [ -inf .. +inf ]
    """

    POSE_ARRAY = "pose_array"
    """
    Pose Array
    **Output when**: Tracking is enabled.
    **Purpose**: Contains the poses defining the bounding box of a tracked object.
    **Format**: 12 floats of 32-bits per object:
    * Tx, Ty and Tz represent the translation vector T to get the position coordinates in meters.
    * Rxx, Rxy, Rxz, Ryx, Ryy, Ryz, Rzx, Rzy and Rzz represent the rotation matrix R defining the orientation.
    As a consequence, considering a device in a pose defined by T and R, you can compute the coordinates Xabs in the
    absolute referential from the coordinates Xrel in device referential using the vectorial formula:
    Xabs = R * Xrel + T
    **Coordinate Frame**:
    Absolute/World Frame, which is:
    - Single-LiDAR configuration: On-Ground LiDAR Coordinate Frame.
    - Multi-LiDARs configuration: Map Coordinate Frame.
    """

    OBJECT_ID = "object_id"
    """
    Object ID [DEPRECATED USAGE]
    **Instead, use**: "Object ID 32 bits" (Type 47)
    """

    CARTESIAN_COORDINATES_4F = "cartesian_coordinates_4f"
    """
    Cartesian Coordinates 4F [DEPRECATED USAGE]
    **Instead, use**: "Cartesian Coordinates" (Type 6)
    Alternative way to represent coordinates, where a fourth float is added.
    It can be more efficient to construct if an application aligns the points to use SIMD on 128 bits words.
    Contains list of cartesian single-precision float coordinates, four per point:
    * x : meters, range [ -inf .. +inf ]
    * y : meters, range [ -inf .. +inf ]
    * z : meters, range [ -inf .. +inf ]
    * w : unused, for 128 bits alignment only
    """

    SPHERICAL_COORDINATES_4F = "spherical_coordinates_4f"
    """
    Spherical Coordinates 4F [DEPRECATED USAGE]
    Alternative way to represent coordinates, where a fourth float is added.
    It can be more efficient to construct if an application aligns the points to use SIMD on 128 bits words.
    Contains list of spherical single-precision float coordinates, four per point:
    * azimuth   : degrees, range [ 0.0 .. 360.0 ]
    * elevation : degrees, range [ -90.0 .. +90.0 ]
    * distance  : meters,  range [ 0.0 ..  +inf ]
    * w         : unused, for 128 bits alignment only
    """

    ZONES_DEF = "zones_def"
    """
    Zones Def
    **Output when**: Tracking is enabled, and at least a zone is defined.
    **Purpose**: Definition of the event zones. They represent spatial areas of interest.
    Their order is important, since the index is used to identify a zone in type 48.
    Contains following sub-TLVs:
    - zero to many:    Zone                           (ZONE)
    """

    ZONE = "zone"
    """
    Zone
    **Output when**: Tracking is enabled, and at least a zone is defined.
    **Purpose**: Defines one zone.
    Contains following sub-TLVs:
    - exactly one:     Zone Vertices                  (ZONE_VERTICES)
    - exactly one:     Zone Name                      (ZONE_NAME)
    - none or one:     Zone UUID                      (ZONE_UUID)
    - none or one:     Zone Vertical Limits           (ZONE_VERTICAL_LIMITS)
    """

    ZONE_VERTICES = "zone_vertices"
    """
    Zone Vertices
    **Output when**: Tracking is enabled, and a zone has been defined.
    **Purpose**: Vertices of the polygon defining the zone.
    They are defined on the ground, so the z coordinate is absent.
    **Format**: Contains list of cartesian single-precision float coordinates, two per point.
    * x : meters, range [ -inf .. +inf ]
    * y : meters, range [ -inf .. +inf ]
    There must be at least 3 vertices, so at least 6 floats.
    """

    ZONE_NAME = "zone_name"
    """
    Zone Name
    **Output when**: Tracking is enabled, and at least a zone is defined.
    **Purpose**: User-defined name to the zone, do not use as unique identifier.
    **Format**: It is UTF-8 encoded and null-terminated.
    """

    _ZONE_UUID = "zone_uuid"
    """
    Zone UUID
    128 bits UUID of the zone in BIG-endian representation.
    UUID 00112233-4455-6677-8899-aabbccddeeff is encoded as the bytes 00 11 22 33 44 55 66 77 88 99 aa bb cc dd ee ff.
    It enables to keep the identity of a zone across renaming or resizing.
    """

    ZONES_OBJECTS_BINDING = "zones_objects_binding"
    """
    Zones Objects Binding [DEPRECATED USAGE]
    **Instead, use**: "Zones Objects Binding 32 bits" (Type 48)
    Concatenation of 0 to N couples of:
    * an unsigned 64-bits integer, representing the ID of an object (see type 30)
    * an unsigned 32-bits integer, representing the index of the zone (the n-th type 34 of type 33)
    Each object-zone couple means that the object is considered by the algorithm to be in the zone.
    """

    OBJECT_PROPERTIES = "object_properties"
    """
    Object Properties
    **Output when**: Tracking is enabled.
    **Purpose**: Properties of the object.
    **Format**: One byte per object, each bit of this can represent different properties.
    * 1st bit: is_oriented: 1 if the object has a proper orientation (like a cuboid), 0 otherwise (like a cylinder).
    * 2nd bit: is_seen: 1 if the object has been seen in the last scan, 0 if it was not seen.
    * 3rd bit: has_valid_slam_pose: 1 if the object has a valid slam pose, 0 otherwise. See type 55 (Slam Pose Array)
    for more information.
    * 4th bit: is_static_object: 1 if the object is static (e.g. traffic signs, poles...), 0 otherwise
    * 5th bit: has_a_priori_dimensions: 1 if an a priori estimation of the object dimensions has been given as input, 0
    otherwise
    * 6th bit: is_controlled: 1 if the object has been declared controlled by a system taking Shift Perception data as
    input
    * 7th bit: reserved for later use
    * 8th bit: reserved for later use
    """

    _IMU_PACKET = "imu_packet"
    """
    IMU Packet
    Contains the following data:
    * Time when the IMU packet was received
    * Byte [0,3]: UNIX time in seconds,   unsigned 32-bits integer, range [ 0 .. 2^32 [
    * Byte [4,7]: remaining microseconds, unsigned 32-bits integer, range [ 0 .. 1000000 [
    * Acceleration vector
    * Byte [8,11] x : meters / second^2, 32 bits float, range [ -inf .. +inf ]
    * Byte [12,15] y : meters / second^2, 32 bits float, range [ -inf .. +inf ]
    * Byte [16,19] z : meters / second^2, 32 bits float, range [ -inf .. +inf ]
    * Angular velocity vector
    * Byte [20,23] x : radians / second, 32 bits float, range [ -inf .. +inf ]
    * Byte [24,27] y : radians / second, 32 bits float, range [ -inf .. +inf ]
    * Byte [28,31] z : radians / second, 32 bits float, range [ -inf .. +inf ]
    """

    _TIMESTAMP_LIDAR_VELODYNE = "timestamp_lidar_velodyne"
    """
    Timestamp Lidar Velodyne
    Describes a timestamp with microsecond precision. This timestamp is in the lidar reference system, and uses the
    Velodyne format. This means that this timestamp is the time spent since the beginning of the hour.
    Contains concatenation of:
    * UNIX time in seconds,   unsigned 32-bits integer, range [ 0 .. 2^32 [
    * remaining microseconds, unsigned 32-bits integer, range [ 0 .. 1000000 [
    """

    POSE_RELATIVE = "pose_relative"
    """
    Pose Relative
    **Output when**: Slam is enabled.
    **Purpose**: This pose represents the movement of the device between two scans.
    So the position of the current scan can be computed from the one of the previous scan using the following vectorial
    formula:
    ```
    Xcurrent = R * Xprevious + T
    ```
    with:
    ```
        [Rxx Rxy Rxz]
    R = [Ryx Ryy Ryz]
        [Rzx Rzy Rzz]
        [Tx]
    T = [Ty]
        [Tz]
    ```
    **Format**: Concatenation of 12 floats of 32-bits each, in this order:
    - Tx, Ty and Tz, representing the translation vector T to get the position coordinates in meters
    - Rxx, Ryx, Rzx, Rxy, Ryy, Rzy, Rxz, Ryz and Rzz, representing the rotation matrix R (given column-wise) defining
    the orientation
    **Coordinate Frame**: relative to the pose of the previous scan.
    """

    _GRAVITY = "gravity"
    """
    Gravity
    Concatenation of 3 floats of 32-bits each [x, y, z], this is the direction of the gravity in the acquisition sensor
    reference frame.
    The vector is either normalised if valid, or [0, 0, 0] if invalid.
    """

    EGO_MOTION = "ego_motion"
    """
    Ego Motion
    **Output when**: SLAM is enabled.
    Contains following sub-TLVs:
    - exactly one:     Pose Relative                  (POSE_RELATIVE)
    - none or one:     Predicted Position             (PREDICTED_POSITION)
    - none or one:     Smoothed Pose                  (SMOOTHED_POSE)
    - none or one:     Divergence Indicator           (DIVERGENCE_INDICATOR)
    - none or one:     Instantaneous Translation Speed (INSTANTANEOUS_TRANSLATION_SPEED)
    - none or one:     Instantaneous Rotation Speed   (INSTANTANEOUS_ROTATION_SPEED)
    - none or one:     Filtered Translation Speed     (FILTERED_TRANSLATION_SPEED)
    - none or one:     Filtered Rotation Speed        (FILTERED_ROTATION_SPEED)
    - none or one:     Cartesian covariance           (CARTESIAN_COVARIANCE)
    - none or one:     Cylindrical covariance         (CYLINDRICAL_COVARIANCE)
    """

    _PREDICTED_POSITION = "predicted_position"
    """
    Predicted Position
    **Output when**: Slam is enabled.
    **Purpose**: Cartesian coordinates of the predicted position in slam.
    The time of prediction is set at startup, and defaults to 1 second.
    **Format**: Contains 3 floats.
    **Coordinate Frame**: relative to the LiDAR Coordinate Frame.
    """

    GEOGRAPHIC_POSE = "geographic_pose"
    """
    Geographic Pose [DEPRECATED USAGE]
    **Deprecated since**: 5.3
    **Instead, use**: "Geographic Pose Precise" (Type 57)
    **Purpose**: Represents the geographic pose, output from the relocalization processing, expressed in decimal
    degrees notation (latitude, longitude & heading).
    **Format**: Concatenation of 3 single-precision (32 bits) floats [lat, long, heading].
    * Single-precision floating-point (float32) : latitude in decimal degrees, range [ -90.0 .. 90.0 ]
    * Single-precision floating-point (float32) : longitude in decimal degrees, range [ -180.0 .. 180.0 ]
    * Single-precision floating-point (float32) : heading in decimal degrees, range [ 0.0 .. 360.0 [
    """

    OBJECT_ID_32_BITS = "object_id_32_bits"
    """
    Object ID 32 Bits
    **Output when**: Tracking is enabled.
    * Standard behavior: output as leaf of Tracked Objects.
    * Optional & internal behavior: output also as leaf of Augmented Cloud when object_id is listed in
    cloud_fields_to_stream.
    **Format**: Array of unsigned 32-bits integers. Each element of the array is an Object ID.
    **Purpose**: Give a unique identifier (ID) to each detected object.
    Optionally, this ID can also be used to establish a link between detected objects
    and the defined Zones, or the points from an Augmented Cloud.
    When used as leaf of Tracked Objects:
    - The number of elements of the array is equal to the Number of Objects (type 10).
    - Each ID represents a distinct detected object.
    - The affectation of an ID to an object is arbitrary and permanent,
      it remains the same throughout the tracking phase.
    When used as leaf of Augmented Cloud:
    - The number of elements of the array is equal to the Number of Points (type 2).
    - Each specified ID indicates to which object the point belongs.
    - The special value 0 is used to mean 'no object'.
      For instance, if the object ID of a point is 0,
      it means that this point does not belong to any object.
    The Object ID is also used in Zones Objects Binding 32 Bits (type 48), combined with Zone identifier.
    """

    ZONES_OBJECTS_BINDING_32_BITS = "zones_objects_binding_32_bits"
    """
    Zones Objects Binding 32 Bits
    **Output when**: Tracking is enabled, and at least one zone has been defined.
    **Purpose**: This type is used to know with tracked objects are in which event zones.
    The binary payload is the concatenation of 0 to N couples of:
    * an unsigned 32-bits integer, representing the ID of an object (see 47)
    * an unsigned 32-bits integer, representing the index of the event zone (the n-th type 34 of type 33)
    Each object-zone couple means that the object is considered by the algorithm to be in the zone.
    """

    _BACKGROUND_BITS = "background_bits"
    """
    Background Bits
    Contains a padded list of bits, 1 bit per point of the cloud. If the bit is set, the point is a background point.
    The background is defined as the objects which were frequently detected in the past.
    On the contrary the foreground is composed of objects that were not seen before (i.e. they appeared recently and/or
    they are moving).
    """

    _GROUND_PLANE_BITS = "ground_plane_bits"
    """
    Ground Plane Bits
    Contains a padded list of bits, 1 bit per point of the cloud. If the bit is set, the point belongs to the ground.
    """

    _AZIMUTHS = "azimuths"
    """
    Azimuths
    Contains list of azimuth coordinates.
    The definition of this angle can be found
    [here](https://docs.outsight.ai/software/coordinate-conventions#lidar-reference-frame).
    The azimuths are in degrees and expressed as single-precision floats. Several values are concatenated so the
    information of n points are given at once.
    The range of each value is [ 0.0 .. 360.0 [. 360.0 [.
    """

    _ELEVATIONS = "elevations"
    """
    Elevations
    Contains list of elevation values in degrees expressed as single-precision floats.
    The definition of the elevation can be found
    [here](https://docs.outsight.ai/software/coordinate-conventions#lidar-reference-frame).
    The range is [ -90.0 .. +90.0 [.
    """

    _DISTANCES = "distances"
    """
    Distances
    Contains list of distance values in meters expressed as single-precision floats.
    The definition of the distance can be found
    [here](https://docs.outsight.ai/software/coordinate-conventions#lidar-reference-frame).
    The range is [ 0.0 .. +inf [.
    """

    _LIDAR_MODEL = "lidar_model"
    """
    LiDAR Model
    One unsigned 8-bits integer representing a model of LiDAR, using enum LIDAR MODEL ID.
    """

    SLAM_POSE_ARRAY = "slam_pose_array"
    """
    Slam Pose Array
    **Output when**: Tracking/objects_super_resolution is enabled.
    **Purpose**: Pose of tracked objects. It is similar to type 29 "Pose Array", but it is more accurate since
    it uses a SLAM algorithm and not just a box fitting algorithm.
    The position part follows an arbitrary point on the object, which is not necessarily its center.
    The orientation part is arbitrary for the first predicted pose, then follows the rotation of the object.
    The SLAM algorithm may not run on all tracked objects. When the SLAM does not run on an object,
    let say on object number 5, all floats of the related pose in this array, e.g. the fifth, are equal to zero.
    Moreover, the related element of type 39 "Object properties", e.g. the fifth, has its third bit equal to zero.
    For further details see: [Spatial Data / Static
    Tracking](https://docs.outsight.ai/software/understanding-osef-data/spatial-data-tracking).
    **Format**: The binary payload is a repetition of the following content, one per tracked object:
    * single-precision floats Tx, Ty and Tz represent the translation vector T to get the position coordinates in
    meters.
    * single-precision floats Rxx, Rxy, Rxz, Ryx, Ryy, Ryz, Rzx, Rzy and Rzz represent the rotation matrix R defining
    the orientation.
    **Coordinate Frame**:
    Absolute/world, which is:
    - Single-LiDAR: On-Ground LiDAR Coordinate Frame.
    - Multi-LiDARs: Map Coordinate Frame.
    """

    ZONE_VERTICAL_LIMITS = "zone_vertical_limits"
    """
    Zone Vertical Limits
    **Output when**: Tracking is enabled, and at least a zone has been defined with Vertical Limits.
    **Format**: The binary value is the concatenation of:
    * one single-precision float, elevation, in meters
    * one single-precision float, height, in meters
    **Purpose**: Optional zone configuration which adds a filtering depending on the vertical position of objects.
    It is part of the definition of the zone.
    Only objects whose altitude is between elevation and elevation + height will be considered in the zone.
    """

    GEOGRAPHIC_POSE_PRECISE = "geographic_pose_precise"
    """
    Geographic Pose Precise
    **Output when**: Relocalization in a reference map and georeferencing are enabled.
    So in processing config: slam.enable set to true, slam.reference_map defined and georeferencing.enable set to true.
    **Purpose**: Represents the geographic pose, output from the relocalization processing, expressed in decimal
    degrees notation (latitude, longitude & heading).
    **Format**: Concatenation of 2 double-precision (64 bits) and a single-precision (32 bits) floats [lat, long,
    heading].
    * Double-precision floating-point (float64) : latitude in decimal degrees, range [ -90.0 .. 90.0 ]
    * Double-precision floating-point (float64) : longitude in decimal degrees, range [ -180.0 .. 180.0 ]
    * Single-precision floating-point (float32) : heading in decimal degrees, range [ 0.0 .. 360.0 [
    """

    _ROAD_MARKINGS_BITS = "road_markings_bits"
    """
    Road Markings Bits [DEPRECATED USAGE]
    Contains a padded (up to a byte) list of bits, 1 bit per point of the cloud. If the bit is set, the point belongs
    to a road marking.
    """

    SMOOTHED_POSE = "smoothed_pose"
    """
    Smoothed Pose
    **Output when**: SLAM is enabled.
    **Purpose**: Smoothed pose of the absolute LiDAR pose will introduce a slight delay.
    **Format**: Concatenation of 12 floats of 32-bits each, see type Pose for more details.
    """

    _INTERACTIVE_REQUEST = "interactive_request"
    """
    Interactive Request
    A Request to a device, where a Timestamp and a Request ID are mandatory.
    """

    _INTERACTIVE_RESPONSE = "interactive_response"
    """
    Interactive Response
    A Response made by a device.
    """

    _INTERACTIVE_REQUEST_ID = "interactive_request_id"
    """
    Interactive Request ID
    An ID to identify to which request is associated the response.
    **Format**: Unsigned 32-bits integer.
    """

    _INTERACTIVE_REQUEST_BACKGROUND_HEADER = "interactive_request_background_header"
    """
    Interactive Request Background Header
    Request background header.
    """

    _INTERACTIVE_RESPONSE_BACKGROUND_HEADER = "interactive_response_background_header"
    """
    Interactive Response Background Header
    Response background header.
    """

    _INTERACTIVE_REQUEST_BACKGROUND_DATA = "interactive_request_background_data"
    """
    Interactive Request Background Data
    Request background data.
    """

    _INTERACTIVE_RESPONSE_BACKGROUND_DATA = "interactive_response_background_data"
    """
    Interactive Response Background Data
    Response background data.
    """

    _INTERACTIVE_REQUEST_PINGPONG = "interactive_request_pingpong"
    """
    Interactive Request PingPong
    Request pingpong.
    **Format**: Binary payload
    """

    _INTERACTIVE_RESPONSE_PINGPONG = "interactive_response_pingpong"
    """
    Interactive Response PingPong
    Response pingpong.
    **Format**: The same binary payload as in the request
    """

    _HEIGHT_MAP = "height_map"
    """
    Height Map
    Height map of a configurable zone of interest.
    **Output when**: generated by height_map app.
    """

    _HEIGHT_MAP_POINTS = "height_map_points"
    """
    Height Map Points
    Points that define the height map and provide the height data.
    The height map is defined as a two-dimensional horizontal grid in the Lidar frame, with given dimensions and number
    of cells.
    The properties of the height map, such as size, number of cells, height offset, etc, are defined by the height_map
    app configuration.
    Every points of the augmented cloud are used to build the height map, with no distinction if it is a point of the
    ground, a motion point or any other flag.
    A point of the cloud is associated with a cell if its projected point on the grid is in the cell.
    The height map provides one height information per cell, the height value depends on the way it is computed by the
    app (a cell with no points will have a nan value).
    The Height Map Points is defined as an array of points, with one point per cell of the height map.
    The x and y coordinates represent the center of the cell in the Lidar frame, the third value is the height of the
    cell computed by the app with the points of the cell.
    **Format**: Contains list of Cartesian single-precision float coordinates, three per point:
    * x : meters, range [ -inf .. +inf ]
    * y : meters, range [ -inf .. +inf ]
    * z : meters, range [ -inf .. +inf ]
    """

    DIVERGENCE_INDICATOR = "divergence_indicator"
    """
    Divergence Indicator
    **Since**: 5.6.0
    **Output when**: Ego-Motion or Relocalization is enabled.
    **Format**: One float value.
    **Purpose**: Indicates if SLAM seems to have diverged.
    - 0 means SLAM algorithm does not seem to have diverged.
    - 1 means SLAM algorithm seems to have diverged, the associated pose is not trustworthy.
    """

    _CARLA_TAG_ARRAY = "carla_tag_array"
    """
    Carla Tag Array
    **Output when**: generated by Carla Scenario Generator.
    **Format**: List of unsigned 8-bits integers, using enum Carla Tag enum.
    **Purpose**: keeps the tag value, attributed by the Carla Simulator, to each point of the point cloud.
    """

    _BACKGROUND_SCENE_PARAMS = "background_scene_params"
    """
    Background Scene Params
    **Output when**: Multi-LiDARS Tracking is enabled, in the background stream between edge and fusion.
    **Purpose**: All params to initialize the background scene and be able to interpret Background Stream Fragment
    (type 76) packets
    """

    _BACKGROUND_SCENE_PARAMS_GENERAL = "background_scene_params_general"
    """
    Background Scene Params General
    **Output when**: Multi-LiDARS Tracking is enabled, in the background stream between edge and fusion.
    **Format**: The following concatenated values:
      * width: 32-bits unsigned integer, number of columns of the background scene array
      * height: 32-bits unsigned integer, number of rows of the background scene array
      * first azimuth: single-precision float, azimuth of the first column of the background scene array, in degrees
      * azimuth step: single-precision float, difference between the azimuths of two consecutive columns of the
    background scene array, in degrees
    """

    _BACKGROUND_SCENE_PARAMS_ELEVATIONS = "background_scene_params_elevations"
    """
    Background Scene Params Elevations
    **Output when**: generated by Carla Scenario Generator.
    **Format**: List of single-precision floats, each one representing the elevation of a row in the background scene
    array, in degrees
    """

    _BACKGROUND_SCENE_FRAGMENT = "background_scene_fragment"
    """
    Background Scene Fragment
    **Output when**: Multi-LiDARS Tracking is enabled, in the background stream between edge and fusion.
    **Purpose**: A fragment of the background scene array
    """

    _BACKGROUND_SCENE_FRAGMENT_INFO = "background_scene_fragment_info"
    """
    Background Scene Fragment Info
    **Output when**: in background stream between edge and fusion, in tracking mode
    **Format**: The following concatenated values:
      * first_index: 32-bites unsigned integer, index of the first cell in background scene array, computed as x *
    height + y, the second cell index is first_index+1, etc., with a rollover on width * height
      * cells_number: 32-bites unsigned integer, number of cells contained in the fragment
    """

    _BACKGROUND_SCENE_FRAGMENT_DISTANCES = "background_scene_fragment_distances"
    """
    Background Scene Fragment Distances
    **Output when**: Multi-LiDARS Tracking is enabled, in the background stream between edge and fusion.
    **Format**: List of single-precision floats, each one representing the farthest background distance for the
    corresponding cell
    """

    GEOGRAPHIC_POSE_ARRAY = "geographic_pose_array"
    """
    Geographic Pose Array
    **Output when**: Tracking is enabled and georeferencing is activated.
    **Purpose**: Represents the geographic pose, output from the relocalization processing, expressed in decimal
    degrees notation (latitude, longitude & heading) of each tracked object.
    **Format**: Concatenation of 2 double-precision (64 bits) and a single-precision (32 bits) floats [lat, long,
    heading], one for each tracked object.
    * Double-precision floating-point (float64): latitude in decimal degrees, range [ -90.0 .. 90.0 ]
    * Double-precision floating-point (float64): longitude in decimal degrees, range [ -180.0 .. 180.0 ]
    * Single-precision floating-point (float32): heading in decimal degrees, range [ 0.0 .. 360.0 [
    """

    GEOGRAPHIC_SPEED = "geographic_speed"
    """
    Geographic Speed
    **Output when**: Georeferencing is activated.
    **Purpose**: Represents speed of mobile platform (on which LiDAR is mounted) as speed in m/s and geographic heading
    in decimal degrees.
    **Format**: Concatenation of two single-precision (32-bits) floats per object:
    * v: absolute speed in meters per second, range [ 0 .. +inf ]
    * heading: speed vector heading in decimal degrees, range [ 0.0 .. 360.0 [
    """

    GEOGRAPHIC_SPEED_ARRAY = "geographic_speed_array"
    """
    Geographic Speed Array
    **Output when**: Tracking is enabled and georeferencing is activated.
    **Purpose**: Represents speed of tracked objects represented as speed in m/s and geographic heading in decimal
    degrees.
    **Format**: Array of geographic speeds, two single-precision (32 bits) floats per object:
    * v: absolute speed in meters per second, range [ 0 .. +inf ]
    * heading: speed vector heading in decimal degrees, range [ 0.0 .. 360.0 [
    """

    _INSTANTANEOUS_TRANSLATION_SPEED = "instantaneous_translation_speed"
    """
    Instantaneous Translation Speed
    **Output when**: SLAM is enabled and its speed_measurement module is enabled.
    **Purpose**: Represents the instantaneous translation speed of the device (m/s).
    **Format**: Array of 3 single-precision (32 bits) floats, representing the translation speed on each direction.
    """

    _INSTANTANEOUS_ROTATION_SPEED = "instantaneous_rotation_speed"
    """
    Instantaneous Rotation Speed
    **Output when**: SLAM is enabled and its speed_measurement module is enabled.
    **Purpose**: Represents the instantaneous rotation speed of the device (rad/s).
    **Format**: Array of 3 single-precision (32 bits) floats, representing the rotation speed around each axis.
    """

    _FILTERED_TRANSLATION_SPEED = "filtered_translation_speed"
    """
    Filtered Translation Speed
    **Output when**: SLAM is enabled and its speed_measurement module is enabled.
    **Purpose**: Represents the filtered translation speed of the device (m/s). An averaging filter smooths the
    evolution of the speed. A short delay will appear due to the filtering (usually around 1 sec).
    **Format**: Array of 3 single-precision (32 bits) floats, representing the translation speed on each direction.
    """

    _FILTERED_ROTATION_SPEED = "filtered_rotation_speed"
    """
    Filtered Rotation Speed
    **Output when**: SLAM is enabled and its speed_measurement module is enabled.
    **Purpose**: Represents the filtered rotation speed of the device (rad/s). An averaging filter smooths the
    evolution of the speed. A short delay will appear due to the filtering (usually around 1 sec).
    **Format**: Array of 3 single-precision (32 bits) floats, representing the rotation speed around each axis.
    """

    REFERENCE_MAP_BITS = "reference_map_bits"
    """
    Reference Map Bits
    **Output when**: SLAM is enabled and its mapping module is enabled.
    **Purpose**: SLAM identifies potential points of interest within the point cloud, and the mapping module selects
    some of them to build the map. These bits identify the points that are selected to generate the output map.
    **Format**: Padded list of bits, 1 bit per point of the cloud. If the bit is set, the point is part of the
    reference map.
    """

    _CARTESIAN_COVARIANCE = "cartesian_covariance"
    """
    Cartesian covariance
    **Output when**: SLAM is enabled and compute_covariance option is set to true.
    **Purpose**: Represents the cartesian covariances (m²). It estimates the confidence of the algorithm, and how
    trustworthy is the position.
    **Format**: Array of 3 single-precision (32 bits) floats, representing the covariance of x, y and z measurements.
    """

    _CYLINDRICAL_COVARIANCE = "cylindrical_covariance"
    """
    Cylindrical covariance
    **Output when**: SLAM is enabled and compute_covariance option is set to true.
    **Purpose**: Represents the cylindrical covariances (°²). It estimates the confidence of the algorithm, and how
    trustworthy is the orientation.
    **Format**: Array of 3 single-precision (32 bits) floats, representing the covariance of roll (rotation around X
    axis), pitch (rotation around Y axis) and yaw (rotation around Z axis) measurements.
    """

    COORDINATES_REFERENCE_SYSTEM = "coordinates_reference_system"
    """
    Coordinates Reference System
    **Output when**: Augmented Cloud is enabled.
    **Purpose**: The coordinates of the augmented cloud may be expressed in an absolute or relative coordinate system.
    The latter usually being relative to the sensor. This field indicates the reference system in use.
    **Format**: Enumerate COORDINATES REFERENCE SYSTEM
    """

    START_TIMESTAMP = "start_timestamp"
    """
    Start timestamp
    **Purpose**: Start of the time interval. The start timestamp is included in the interval. Unix timestamp with
    **nanosecond** precision.
    **Format**: Contains concatenation of:
    * UNIX time in seconds, unsigned 32-bits integer, range [ 0 .. 2^32 [
    * remaining nanoseconds, unsigned 32-bits integer, range [ 0 .. 1'000'000'000 [
    """

    END_TIMESTAMP = "end_timestamp"
    """
    End timestamp
    **Purpose**: End of the time interval. The end timestamp is not included in the interval. Unix timestamp with
    **nanosecond** precision.
    **Format**: Contains concatenation of:
    * UNIX time in seconds, unsigned 32-bits integer, range [ 0 .. 2^32 [
    * remaining nanoseconds, unsigned 32-bits integer, range [ 0 .. 1'000'000'000 [
    """

    TIME_DOMAIN = "time_domain"
    """
    Time domain
    **Purpose**: Time domain used to generate the timestamps:
       * Reception time: timestamps are generated when the data is received by the system,
       * Sensor time: timestamps are generated by the LiDAR, when the data is acquired
    Warning: Mixing timestamps of different domains is not advised.
    **Format**: Unsigned 32-bits enumerate TIME DOMAIN
    """

    TIME_INTERVAL = "time_interval"
    """
    Time interval
    **Output when**: Scan Maker is time-based.
    **Purpose**: A duration between a start timestamp and an end timestamp. Depending on the time domain Sub-TLV, the
    data is guaranteed to have been either acquired or received during this duration.
    Contains following sub-TLVs:
    - exactly one:     Start timestamp                (START_TIMESTAMP)
    - exactly one:     End timestamp                  (END_TIMESTAMP)
    - exactly one:     Time domain                    (TIME_DOMAIN)
    """

    CLASSIFICATION_CONFIDENCE = "classification_confidence"
    """
    Classification confidence
    **Output when**: Tracking is enabled.
    **Purpose**: Represents a level of confidence, from 0 to 1, regarding the class ID assigned to each tracked object.
    **Format**: A single-precision (32 bits) float for each tracked object, range [ 0 .. 1 ].
    """

    _CLASS_PROBABILITIES_ARRAY = "class_probabilities_array"
    """
    Class probabilities Array
    **Output when**: Tracking and class probabilities are enabled.
    **Purpose**: Normalized distribution over enabled class IDs, expressing, for each tracked object, its
    classification probability.
    **Format**: Array containing a single-precision (32 bits) float per class id for each tracked object with range [ 0
    .. 1 ].
    """

    _CLASSIFIER_DEF = "classifier_def"
    """
    Classifier Def
    **Output when**: Tracking is enabled.
    **Purpose**: Definition of classifier settings
    Contains following sub-TLVs:
    - exactly one:     Enabled class IDs Array        (ENABLED_CLASS_IDS_ARRAY)
    """

    _ENABLED_CLASS_IDS_ARRAY = "enabled_class_ids_array"
    """
    Enabled class IDs Array
    **Output when**: Tracking and class probabilities are enabled.
    **Purpose**: List class IDs that can be assigned to tracked objects.
    **Format**: Array of signed 32 bits integers, using enum CLASS ID. Use TLV length to determine the number of class
    IDs.
    The order is important as it is respected for other fields like type 95.
    """

    _REFLECTION_FEATURES_ARRAY = "reflection_features_array"
    """
    Reflection features array
    **Output when**: Tracking is enabled and option is enabled in configuration
    **Purpose**: Internal only: Reflections filtering - values of features for each object used by the reflection
    filtering. Used for debug and/or retraining.
    **Format**: Array of single precision floats, 6 per objects, concatenated for all objects.
    """

    _REFLECTION_BITS = "reflection_bits"
    """
    Reflection Bits
    Contains a padded list of bits, 1 bit per point of the cloud. If the bit is set, the point is an unwanted
    reflection, generated by ricochet on an overly reflective surface.
    """


class ClassId(IntEnum):
    """Tracked object class ID"""

    UNKNOWN = 0
    PERSON = 1
    LUGGAGE = 2
    TROLLEY = 3
    DEPRECATED = 4
    TRUCK = 5
    CAR = 6
    VAN = 7
    TWO_WHEELER = 8
    MASK = 9
    NO_MASK = 10
    LANDMARK = 11
    TRAILER = 12
    TRACTOR_HEAD = 13
    GATE = 14
    TUGGER_TRAIN = 15


class PerceptId(IntEnum):
    """Percept class ID"""

    DEFAULT = 0
    ROAD = 1
    VEGETATION = 2
    GROUND = 3
    SIGN = 4
    BUILDING = 5
    FLAT_GND = 6
    UNKNOWN = 7
    MARKING = 8
    OBJECT = 9
    WALL = 10


class LidarModelId(IntEnum):
    """Lidar model ID"""

    UNKNOWN = 0
    VELODYNE_VLP16 = 1
    VELODYNE_VLP32 = 2
    VELODYNE_VLS128 = 3
    VELODYNE_HDL32 = 4
    ROBOSENSE_BPEARL_V1 = 5
    ROBOSENSE_BPEARL_V2 = 6
    ROBOSENSE_RS32 = 7
    ROBOSENSE_HELIOS = 8
    LIVOX_HORIZON = 9
    LIVOX_AVIA = 10
    LIVOX_MID70 = 11
    OUSTER = 12
    OUTSIGHT_SA01 = 13
    HESAI_PANDARXT32 = 14
    HESAI_PANDARQT64 = 15
    FAKE_LIDAR = 16
    HESAI_PANDARXT32M2X = 17
    ROBOSENSE_M1 = 18
    HESAI_PANDAR128E3X = 19
    LEISHEN_C32A = 20
    LEISHEN_C32C = 21
    LIVOX_HAP_TX = 22
    INNOVIZ_ONE_EAGLE = 23
    INNOVIZ_ONE_FALCON = 24
    HESAI_PANDAR_64 = 25
    HESAI_QT128 = 26
    INNOVUSION_FALCON_PRIME = 27
    LIVOX_MID360 = 28
    ROBOSENSE_BPEARL_V3 = 29
    ROBOSENSE_BPEARL_V4 = 30
    HESAI_PANDAR128E4X = 31


class CarlaTag(IntEnum):
    """Carla Tag"""

    UNLABELED = 0
    BUILDING = 1
    FENCE = 2
    OTHER = 3
    PEDESTRIAN = 4
    POLE = 5
    ROADLINE = 6
    ROAD = 7
    SIDEWALK = 8
    VEGETATION = 9
    VEHICLES = 10
    WALL = 11
    TRAFFICSIGN = 12
    SKY = 13
    GROUND = 14
    BRIDGE = 15
    RAILTRACK = 16
    GUARDRAIL = 17
    TRAFFICLIGHT = 18
    STATIC = 19
    DYNAMIC = 20
    WATER = 21
    TERRAIN = 22


class CoordinatesReferenceSystem(IntEnum):
    """Coordinates Reference System"""

    UNDEFINED = 0
    WORLD = 1
    SENSOR = 2


class TimeDomain(IntEnum):
    """Time domain"""

    LOCAL = 0
    SENSOR = 1



"""Packing/Unpacking functions and classes for OSEF types."""

# Standard imports
import uuid
from struct import Struct
from typing import Callable, List, Optional

# Third party imports
import numpy as np
from numpy import typing as npt

# Project imports
from osef.spec import constants


class PackerBase:
    """Base class to define a packer.

    Each new packer should derive from this interface.
    In the child classes, assign in the constructor values for packer and unpacker functions,
    so they won't be evaluated/defined again.

    :param pack: Callable function to pack the data (Optional).
    :param unpack: Callable function to unpack the data (Optional).
    """

    def __init__(self) -> None:
        """Constructor."""

        # Init the pack and unpack callable as None.
        self.pack: Optional[Callable] = None
        self.unpack: Optional[Callable] = None


class ValuePacker(PackerBase):
    """Derived class to pack/unpack value."""

    def __init__(self, pack_format: str) -> None:
        """Constructor.

        :param pack_format:
        """
        super().__init__()
        self.pack = ValuePacker._get_value_packer(pack_format)
        self.unpack = ValuePacker._get_value_unpacker(pack_format)

    @staticmethod
    def _get_value_packer(pack_format: str) -> Callable:
        def _pack_value(value: object) -> bytes:
            return Struct(pack_format).pack(value)

        return _pack_value

    @staticmethod
    def _get_value_unpacker(pack_format: str) -> Callable:
        def _parse_value(value: bytes) -> object:
            return (Struct(pack_format).unpack(value))[0]

        return _parse_value


class ArrayPacker(PackerBase):
    """Derived class to pack/unpack array."""

    def __init__(self, dtype: np.dtype, cols: int = 0) -> None:
        """Constructor.

        :param dtype: Dtype of the numpy array.
        :param cols: Number of columns, default to 0.
        """
        super().__init__()
        self.pack = ArrayPacker._get_array_packer(dtype, cols)
        self.unpack = ArrayPacker._get_array_unpacker(dtype, cols)

    @staticmethod
    def _get_array_packer(dtype: np.dtype, cols: int = 0) -> Callable:
        def _array_packer(value: npt.NDArray) -> bytes:
            if value.dtype != dtype:
                raise TypeError(
                    f"Invalid array type to pack: expected = {dtype}, input = {value.dtype}"
                )
            if cols > 0 and value.shape[1] != cols:
                raise TypeError(
                    f"Invalid array cols to pack: expected = {cols}, input = {value.shape[1]}"
                )
            return value.tobytes()

        return _array_packer

    @staticmethod
    def _get_array_unpacker(dtype: np.dtype, cols: int = 0) -> Callable:
        def _parse_array(value: bytes) -> npt.NDArray[dtype]:
            array = np.frombuffer(value, dtype=dtype)
            if cols > 0:
                array = np.reshape(array, (int(array.shape[0] / cols), cols))
            return array

        return _parse_array


class StructuredArrayPacker(PackerBase):
    """Derived class to pack/unpack structured array."""

    def __init__(self, dtype: np.dtype) -> None:
        """Constructor.

        :param dtype: Dtype of the numpy array.
        :param cols: Number of cols, default to 0.
        """
        super().__init__()
        self.pack = ArrayPacker._get_array_packer(dtype)
        self.unpack = StructuredArrayPacker._get_structured_array_unpacker(dtype)

    @staticmethod
    def _get_structured_array_unpacker(dtype: np.dtype) -> Callable:
        def _parse_structured_array(value: bytes) -> npt.NDArray[dtype]:
            array = np.frombuffer(value, dtype=dtype)
            names = array.dtype.names
            if constants.ExtraOsefKeys._TO_DROP in names:
                names.remove(constants.ExtraOsefKeys._TO_DROP)
                array = array[names]
            return array

        return _parse_structured_array


class BytesPacker(PackerBase):
    """Derived class to pack/unpack bytes."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()

        # Packer/Unpacker for bytes just return the same value.
        self.pack = lambda value: value
        self.unpack = lambda value: value


class DictPacker(PackerBase):
    """Derived class to pack/unpack dictionary."""

    def __init__(self, pack_format: str, field_names: List[str]) -> None:
        """Constructor.

        :param pack_format: Format of the dict values.
        :param field_names: Names of the dict fields/keys.
        """
        super().__init__()
        self.pack = DictPacker._get_dict_packer(pack_format, field_names)
        self.unpack = DictPacker._get_dict_unpacker(pack_format, field_names)

    @staticmethod
    def _get_dict_packer(pack_format: str, fields_names: List[str]) -> Callable:
        def _pack_dict(value: dict) -> bytes:
            values = [value[k] for k in fields_names]
            array = Struct(pack_format).pack(*values)
            return array

        return _pack_dict

    @staticmethod
    def _get_dict_unpacker(pack_format: str, fields_names: List[str]) -> Callable:
        def _parse_dict(value: bytes) -> dict:
            array = list(Struct(pack_format).iter_unpack(value))
            return dict(zip(fields_names, array[0]))

        return _parse_dict


class TimestampPacker(PackerBase):
    """Derived class to pack/unpack timestamp."""

    def __init__(self, sub_seconds_order: int) -> None:
        """
        Constructor.
        `sub_seconds_order` is the order (power of 10) associated to the time decimals.
        """
        super().__init__()
        self.sub_seconds_order = sub_seconds_order
        self.pack = self._timestamp_packer
        self.unpack = self._timestamp_unpacker

    def _timestamp_packer(self, value: float) -> bytes:
        seconds = int(value)
        sub_seconds = int(round((value - seconds) * 10**self.sub_seconds_order))
        return Struct("<LL").pack(seconds, sub_seconds)

    def _timestamp_unpacker(self, value: bytes) -> float:
        seconds, sub_seconds = Struct("<LL").unpack(value)
        return seconds + sub_seconds * 10**-self.sub_seconds_order


class TimestampMicroPacker(TimestampPacker):
    """Derived class to pack/unpack timestamp using microseconds."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__(6)


class TimestampNanoPacker(TimestampPacker):
    """Derived class to pack/unpack timestamp using nanoseconds."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__(9)


class ProcessingBitfieldPacker(PackerBase):
    """Derived class for pack/unpack processing bitfield."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.pack = ProcessingBitfieldPacker._processing_bitfield_packer
        self.unpack = ProcessingBitfieldPacker._processing_bitfield_unpacker

    @staticmethod
    def _processing_bitfield_packer(value: dict) -> bytes:
        return Struct("<Q").pack(value[constants.ExtraBackgroundKeys.DELETED])

    @staticmethod
    def _processing_bitfield_unpacker(value: bytes) -> dict:
        background_deleted = 0
        bitfield = Struct("<Q").unpack(value)[0]
        return {
            constants.ExtraBackgroundKeys.DELETED: (
                bitfield & (1 << background_deleted) != 0
            )
        }


class BoolBitfieldPacker(PackerBase):
    """Derived class for pack/unpack bool bitfield."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.pack = BoolBitfieldPacker._bool_bitfield_packer
        self.unpack = BoolBitfieldPacker._bool_bitfield_unpacker

    @staticmethod
    def _bool_bitfield_packer(value: npt.NDArray[np.bool_]) -> bytes:
        return np.packbits(value, bitorder="little").tobytes()

    @staticmethod
    def _bool_bitfield_unpacker(value: bytes) -> npt.NDArray[np.bool_]:
        if len(value) == 0:
            return np.array([], dtype=np.bool_)
        np_8bit = np.frombuffer(value, dtype=np.uint8)
        return np.unpackbits(np_8bit, bitorder="little").astype(np.bool_)


class StringPacker(PackerBase):
    """Derived class for pack/unpack string."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.pack = StringPacker._string_packer
        self.unpack = StringPacker._string_unpacker

    @staticmethod
    def _string_packer(value: str) -> bytes:
        return value.encode("ascii") + b"\x00"

    @staticmethod
    def _string_unpacker(value: bytes) -> str:
        return value.decode("ascii")[:-1]


class ImuPacker(PackerBase):
    """Derived class for pack/unpack IMU data."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.pack = ImuPacker._imu_packer
        self.unpack = ImuPacker._imu_unpacker

    @staticmethod
    def _imu_packer(value: dict) -> bytes:
        return Struct("<LLffffff").pack(
            value[constants.ExtraTimestampKeys.TIMESTAMP][
                constants.ExtraTimestampKeys.UNIX_S
            ],
            value[constants.ExtraTimestampKeys.TIMESTAMP][
                constants.ExtraTimestampKeys.REMAINING_US
            ],
            value[constants.ExtraImuKeys.ACCELERATION][0],
            value[constants.ExtraImuKeys.ACCELERATION][1],
            value[constants.ExtraImuKeys.ACCELERATION][2],
            value[constants.ExtraImuKeys.ANGULAR_VELOCITY][0],
            value[constants.ExtraImuKeys.ANGULAR_VELOCITY][1],
            value[constants.ExtraImuKeys.ANGULAR_VELOCITY][2],
        )

    @staticmethod
    def _imu_unpacker(value: bytes) -> dict:
        value = Struct("<LLffffff").unpack(value)
        return {
            constants.ExtraTimestampKeys.TIMESTAMP: {
                constants.ExtraTimestampKeys.UNIX_S: value[0],
                constants.ExtraTimestampKeys.REMAINING_US: value[1],
            },
            constants.ExtraImuKeys.ACCELERATION: value[2:5],
            constants.ExtraImuKeys.ANGULAR_VELOCITY: value[5:8],
        }


class UuidPacker(PackerBase):
    """Derived class"""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.pack = UuidPacker._uuid_packer
        self.unpack = UuidPacker._uuid_unpacker

    @staticmethod
    def _uuid_packer(value: uuid.UUID) -> bytes:
        return value.bytes

    @staticmethod
    def _uuid_unpacker(value: bytes) -> uuid.UUID:
        return uuid.UUID(bytes=value)


class ObjectPropertiesPacker(PackerBase):
    """Derived class for pack/unpack object properties."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.pack = ObjectPropertiesPacker._object_properties_packer
        self.unpack = ObjectPropertiesPacker._object_properties_unpacker

    @staticmethod
    def _object_properties_packer(object_properties: npt.NDArray[np.bool_]) -> bytes:
        return (
            np.packbits(object_properties, bitorder="little").astype(np.uint8).tobytes()
        )

    @staticmethod
    def _object_properties_unpacker(value: bytes) -> npt.NDArray[np.bool_]:
        array = np.frombuffer(value, dtype=np.dtype(np.ubyte))

        # Use [:, None] trick to avoid to get a flatten array
        return np.unpackbits(array[:, None], axis=1, bitorder="little").astype(bool)



"""Functions to read and parse osef files/streams."""
from __future__ import annotations
from abc import ABC, abstractmethod
import pathlib
import socket
import time
from typing import Union, Optional
from io import BufferedReader
import urllib

from osef._logger import osef_logger

# Constants
TCP_SCHEME = "tcp"
TCP_TIMEOUT = 3


class OsefConnectionException(Exception):
    """Exception raised when failing to connect to a tcp stream/failing to open a file."""


def path_to_str(path: Union[str, pathlib.Path]) -> str:
    """If path is a pathlib.Path instance, convert to str"""
    if isinstance(path, pathlib.Path):
        return str(path.absolute())
    return path


class OsefStream(ABC):
    """Context manager class to open file path or tcp socket, then read its values.
    Delegate actual implementation to child classes
    """

    @abstractmethod
    def __enter__(self) -> OsefStream:
        """Context manager interface. Implementation is delegated to the child class."""

    @abstractmethod
    def __exit__(self, *_) -> None:
        """Context manager interface. Implementation is delegated to the child class."""

    @abstractmethod
    def read(self, _: int) -> bytes:
        """read interface. Implementation is delegated to the child class."""

    @abstractmethod
    def connect(self) -> None:
        """read interface. Implementation is delegated to the child class."""

    @abstractmethod
    def disconnect(self) -> None:
        """read interface. Implementation is delegated to the child class."""


def create_osef_stream(
    path: Union[str, pathlib.Path], auto_reconnect: bool = True
) -> OsefStream:
    """
    Factory for creating an OsefStream
    Instantiate an OsefStreamTcp if given path is tcp, else an OsefStreamFile
    TCP stream if path has form `tcp://hostname:port`.

    :param path: path to osef file or TCP stream.
    :param auto_reconnect: activate auto reconnection in case of connection loss (tcp only)
    """

    if urllib.parse.urlparse(path_to_str(path)).scheme == TCP_SCHEME:
        return OsefStreamTcp(path, auto_reconnect)

    return OsefStreamFile(path)


class OsefStreamTcp(OsefStream):
    """Context manager class to open tcp socket, then read its values.
    TCP stream if path has form `tcp://hostname:port`.

    :param path: path to osef file or TCP stream.
    The server may close the socket if client is too late.
    """

    def __init__(self, path: Union[str, pathlib.Path], auto_reconnect: bool = True):
        """Constructor.
        TCP stream with form `tcp://hostname:port`.

        :param path: TCP stream.
        :param auto_reconnect: activate auto reconnection in case of connection loss
        """
        self._path: urllib.parse.ParseResult = urllib.parse.urlparse(path_to_str(path))
        self._tcp_socket: Optional[socket.socket] = None
        self._auto_reconnect = auto_reconnect

    def __enter__(self) -> OsefStreamTcp:
        """Context manager"""
        self.open_socket(self._auto_reconnect)
        return self

    def __exit__(self, *_) -> None:
        """Context manager"""
        self.disconnect()

    def read(self, size: int = 4096) -> bytes:
        """Read `size` bytes from socket.

        :param size: of binary value to be read
        :raises EOFError: if no value can be read or if it is empty.
        :raises OsefConnectionException: If no TCP socket.
        :return: Read binary value
        """
        msg = b""
        if self._tcp_socket is None:
            raise OsefConnectionException(
                f"TCP socket ({self._path.netloc}) has not been initialized"
            )

        try:
            msg = self._tcp_socket.recv(size)
        except socket.timeout:
            osef_logger.warning(
                f"Received timeout. Closing socket ({self._path.netloc})."
            )
            self.disconnect()
        except ConnectionResetError:
            osef_logger.warning(
                f"Socket reset error. Closing socket ({self._path.netloc})."
            )
            self.disconnect()
        return msg

    def connect(self) -> None:
        """Try to open the socket

        :raises OsefConnectionException: if the connection failed.
        """
        try:
            self._tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._tcp_socket.settimeout(TCP_TIMEOUT)
            self._tcp_socket.connect((self._path.hostname, self._path.port or 11120))
            osef_logger.warning(
                "Connected to %s:%d",
                self._path.hostname,
                self._path.port or 11120,
            )
        except OSError as exp:
            raise OsefConnectionException(f"{exp} ({self._path.netloc})") from exp

    def disconnect(self) -> None:
        """Close the socket"""
        if self._tcp_socket is not None:
            self._tcp_socket.close()

    def open_socket(
        self,
        auto_reconnect: bool = True,
        reconnection_frequency_hz: float = 200,
        log_frequency_hz: float = 2,
    ) -> None:
        """Open tcp socket on provided path.
        Tries to connect again if the connection fails
        """

        if reconnection_frequency_hz < 0:
            raise ValueError(
                f"Invalid reconnection frequency {reconnection_frequency_hz} Hz, should be strictly positive."
            )
        if log_frequency_hz > reconnection_frequency_hz:
            raise ValueError(
                f"Invalid log frequency {log_frequency_hz} Hz,"
                f" should be smaller than the reconnection frequency ({reconnection_frequency_hz} Hz)."
            )

        # Count the retries, to avoid flooding the log
        retry = 0
        waiting_time = 1 / reconnection_frequency_hz
        nb_iteration_between_logs = reconnection_frequency_hz / log_frequency_hz

        while True:
            try:
                self.connect()
                break
            except OsefConnectionException as exp:
                if not retry % nb_iteration_between_logs:
                    osef_logger.error(exp)

            if not auto_reconnect:
                break
            if not retry % nb_iteration_between_logs:
                osef_logger.warning(
                    "Retrying to connect to %s:%d",
                    self._path.hostname,
                    self._path.port or 11120,
                )
            retry = retry + 1
            time.sleep(waiting_time)


class OsefStreamFile(OsefStream):
    """Context manager class to open an osef file, then read its values."""

    def __init__(self, path: Union[str, pathlib.Path]):
        """
        Constructor
        :param path: path to osef file
        """
        self._path: str = path_to_str(path)
        self._io_stream: Optional[BufferedReader] = None

    def connect(self) -> None:
        """Open the file

        :raises OsefConnectionException: if the file could not be opened.
        """
        try:
            self._io_stream = open(  # pylint: disable=consider-using-with
                self._path, "rb"
            )
        except OSError as exp:
            raise OsefConnectionException(f"{exp} ({self._path})") from exp

    def disconnect(self) -> None:
        """Close the file"""
        if self._io_stream is not None:
            self._io_stream.close()

    def __enter__(self) -> OsefStreamFile:
        """Context manager"""
        self.connect()
        return self

    def __exit__(self, *_) -> None:
        """Context manager"""
        self.disconnect()

    def read(self, size: int = 4096) -> bytes:
        """Read `size` bytes from file.

        :param size: of binary value to be read
        :raises EOFError: if no value can be read or if it is empty.
        :raises OsefConnectionException: IOStream not opened.
        :return: Read binary value
        """

        if self._io_stream is None:
            raise OsefConnectionException(
                f"File IOStream has not been opened ({self._path})"
            )

        return self._io_stream.read(size)



"""Functions to lazy read and parse osef files/streams."""

# Standard imports
from typing import Any, Tuple, NamedTuple, List, Optional

# Project imports
from osef.spec import _types as osef_types
from osef.spec import _packers
from osef.spec.constants import _TreeNode
from osef.parsing._parser_common import unpack_value


class BinaryLeaf(NamedTuple):
    """
    Wrapper for binary leaves.
    """

    osef_type: osef_types.OsefTypes
    leaf_value: bytes

    def __repr__(self) -> str:
        """High-level representation of a BinaryLeaf to avoid displaying all bytes of the leaf."""
        return "\x1B[3m(binary blob)\x1B[0m"


def _parse_osef_binary(value: BinaryLeaf) -> Any:
    """
    parse osef binary data

    :param: the binary leaf to be parsed
    :return: the unpacked value
    """

    type_info = osef_types.get_type_info_by_id(value.osef_type)
    return unpack_value(value.leaf_value, type_info.node_info, type_info.name)


class _LazyContainer:
    """Parent class to all lazy containers"""

    def __init__(self) -> None:
        self._unpacking_active: bool = True

    @property
    def unpacking(self) -> bool:
        """Whether unpacking is active or not"""
        return self._unpacking_active

    @unpacking.setter
    def unpacking(self, active: bool) -> None:
        """Switch on/off lazy unpacking (useful when packing)"""
        self._unpacking_active = active


class _LazyDict(dict, _LazyContainer):
    """
    dict with lazy [] operator
    data are stored in binary, and only parsed when the getter is called the first time.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        _LazyContainer.__init__(self)

    def __getitem__(self, key: str) -> Any:
        """
        Override [] dict method with lazy parsing of the data
        The first time a data is accessed (ie it is still binary),
        it is parsed and the parsed value in written in place

        :param key: the key to search for in the dict
        :return Parsed value
        """

        value = super().__getitem__(key)
        # Only case we do something: it's a leaf AND it has not been parsed yet
        if self._unpacking_active and isinstance(value, BinaryLeaf):
            parsed_value = _parse_osef_binary(value)
            # The trick here is that [] in writing mode calls __setitem__
            # Any call to [] as __getitem__ would yield an endless recursion loop
            self[key] = parsed_value
            return parsed_value

        return value

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Override get dict method with lazy parsing of the data
        The first time a data is accessed (ie it is still binary),
        it is parsed and the parsed value in written in place

        :param key: the key to search for in the dict
        :param default: default value if the key does not exist
        :return Parsed value
        """

        if key in self:
            return self[key]
        return default

    def items(self) -> List[Tuple[str, Any]]:
        """
        Override items dict method to return parsed (key, value) tuple.

        :return (key, parsed_value) tuple list
        """

        return [(key, self[key]) for key, _ in super().items()]

    def values(self) -> List[Any]:
        """
        Override values dict method to return parsed value list.

        :return Parsed value list
        """

        return [self[key] for key, _ in super().items()]


class _LazyList(list, _LazyContainer):
    """
    list with lazy [] operator
    data are stored in binary, and only parsed when the getter is called the first time.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _LazyContainer.__init__(self)

    def __getitem__(self, index: int) -> Any:
        """
        Override [] list method with lazy parsing of the data
        The first time a data is accessed (ie it is still binary),
        it is parsed and the parsed value in written in place

        :param index: the index to search for in the list
        :return Parsed value
        """

        value = super().__getitem__(index)
        # Only case we do something: it's a leaf AND it has not been parsed yet
        if self._unpacking_active and isinstance(value, BinaryLeaf):
            parsed_value = _parse_osef_binary(value)
            # The trick here is that [] in writing mode calls __setitem__
            # Any call to [] as __getitem__ would yield an endless recursion loop
            self[index] = parsed_value
            return parsed_value

        return value

    def get(self, index: int, default: Optional[Any] = None) -> Any:
        """
        Override get list method with lazy parsing of the data
        The first time a data is accessed (ie it is still binary),
        it is parsed and the parsed value in written in place

        :param index: the index to search for in the list
        :param default: default value if the key does not exist
        :return Parsed value
        """
        if index in self:
            return self[index]
        return default


def _parse_tree_to_bin_tuple(raw_tree: _TreeNode) -> Tuple[str, Any]:
    """
    Parse a raw TLV tree and, recursively, build a dictionary where leaves are stored as BinaryLeaf
    """
    osef_type, children, leaf_value = raw_tree

    # Get leaf type info
    type_info = osef_types.get_type_info_by_id(osef_type)

    # For leaves, return a BinaryLeaf tuple which holds the binary blob and the osef type
    if isinstance(type_info.node_info, _packers.PackerBase):
        return (
            type_info.name,
            BinaryLeaf(osef_type, leaf_value),
        )

    # For non-leaves, add each child to a dictionary
    tree_bin = _LazyList() if type_info.node_info.node_type == list else _LazyDict()

    for child in children:
        child_name, child_tree_bin = _parse_tree_to_bin_tuple(child)

        if type_info.node_info.node_type == dict:
            tree_bin[child_name] = child_tree_bin
        elif type_info.node_info.node_type == list:
            tree_bin.append(_LazyDict({child_name: child_tree_bin}))
        else:
            raise ValueError("Unsupported internal node type.")

    return type_info.name, tree_bin


def parse_to_lazy_dict(binary_tlv_tree: _TreeNode):
    """Parse a whole frame tree to a _LazyDict.
    No value of the tree will be unpacked until direct access.

    :param binary_tlv_tree: raw tree of a tlv frame.
    :return: _LazyDict with all values in BinaryLeaf.
    """
    type_name, subtree_bin = _parse_tree_to_bin_tuple(binary_tlv_tree)
    return _LazyDict({type_name: subtree_bin})



"""Functions to read and parse osef files/streams."""
import pathlib
import time
from typing import Any, Optional, Iterator, Union, List, Dict

from osef.spec import _types
from osef.spec.osef_types import OsefKeys
from osef._logger import osef_logger
from osef.parsing.osef_stream import (
    OsefConnectionException,
    OsefStream,
    OsefStreamTcp,
    create_osef_stream,
)
from osef.parsing._parser_common import (
    get_tlv_iterator,
    read_next_tlv,
    MalformedTlvException,
    OsefParsingException,
    build_tree,
    _parse_tlv_from_blob,
    _TreeNode,
    _Tlv,
)

from osef.parsing._lazy_parser import parse_to_lazy_dict
from osef.parsing._parser import parse_to_dict, _parse_raw_to_tuple


def parse_timestamp(tlv: _Tlv) -> float:
    """
    Get the timestamp of a frame from its TLV.

    :param tlv: raw TLV blob of a frame.
    :raises ValueError: the frame doesn't contain any timestamp leaf.
    :return: timestamp value as a float
    """
    if tlv.osef_type == _types.OsefTypes.TIMESTAMPED_DATA.value:
        read = 0
        sub_size = 0
        while read < tlv.length:
            sub_tlv, sub_size = _parse_tlv_from_blob(tlv.value, read)
            if sub_tlv.osef_type == _types.OsefTypes.TIMESTAMP_MICROSECOND.value:
                res = _parse_raw_to_tuple(
                    _TreeNode(sub_tlv.osef_type, None, sub_tlv.value)
                )
                return res[1]
            read += sub_size
    raise ValueError("No timestamp found in frame.")


def get_frame_timestamps(
    path: Union[str, pathlib.Path],
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> List[float]:
    """
    Parse the frame timestamps of an OSEF file without unpacking all the frame content.
    If timestamp is missing in a frame, the list will contain timestamps of the previous frames.

    :param path: Path of OSEF file.
    :param first: iterate only on N first frames of file.
    :param last: iterate only on M last frames of file.
        Can be used with first to get the range [(N-M), N].
    :return: List of timestamps as floats of each frame of the file.
    """
    timestamps = []
    with create_osef_stream(path) as osef_stream:
        iterator = get_tlv_iterator(osef_stream, first, last)
        for _, tlv in iterator:
            try:
                timestamps.append(parse_timestamp(tlv))
            except ValueError:
                osef_logger.warning(
                    "A frame without any timestamp has been found."
                    " List of timestamps will be stopped just before this frame."
                )
                break
        return timestamps


# pylint: disable=too-many-locals
def parse(
    path: Union[str, pathlib.Path],
    first: Optional[int] = None,
    last: Optional[int] = None,
    auto_reconnect: bool = True,
    real_frequency: bool = False,
    lazy: bool = True,
) -> Iterator[Dict[str, Any]]:
    """Iterator that opens and converts each tlv frame to a dict.
    Factorization for both legacy parse and lazy parse.
    TCP stream if path has form `tcp://hostname:port`.

    :param path: path to osef file or TCP stream.
    :param first: iterate only on N first frames of file.
    :param last: iterate only on M last frames of file.
        Can be used with first to get the range [(N-M), N].
    :param auto_reconnect: enable reconnection for tcp connections.
    :param real_frequency: processing is slowed at the same frequency as the osef file
    :param lazy: Flag for lazy parsing
    :return: next tlv dictionary
    """

    def get_timestamp_ms(cur_frame_dict: dict) -> int:
        return cur_frame_dict[OsefKeys.TIMESTAMPED_DATA.value][
            OsefKeys.TIMESTAMP_MICROSECOND.value
        ]

    first_frame_timestamp = None
    while True:  # To reopen stream when Tlv is malformed (in auto reconnect mode)
        with create_osef_stream(path, auto_reconnect) as osef_stream:
            first_frame_counter = time.perf_counter()
            iterator = get_tlv_iterator(osef_stream, first, last, auto_reconnect)
            for _, tlv in iterator:
                try:
                    raw_tree = build_tree(tlv)
                except MalformedTlvException as exception:
                    osef_logger.error(f"{exception} (closing OSEF stream)")
                    break  # we disconnect OSEF stream as our reading
                    # head may not be at the beginning of the next frame

                frame_dict = (
                    parse_to_lazy_dict(raw_tree) if lazy else parse_to_dict(raw_tree)
                )

                if first_frame_timestamp is None:
                    first_frame_timestamp = get_timestamp_ms(frame_dict)
                yield frame_dict

                if real_frequency and not isinstance(osef_stream, OsefStreamTcp):
                    _sleep(
                        first_frame_counter,
                        first_frame_timestamp,
                        get_timestamp_ms(frame_dict),
                    )
            if not isinstance(osef_stream, OsefStreamTcp) or not auto_reconnect:
                break


def parse_next_frame(
    osef_stream: OsefStream,
    lazy: bool = True,
) -> Dict[str, Any]:
    """
    Try to parse the next frame. Assume the stream is already connected.

    :param osef_stream: the OsefStream we want to parse data from
    :param lazy: whether to use lazy-parsing
    :raises OsefParsingException: if the next frame could not be parsed.
    :return: next tlv dictionary

    """
    try:
        tlv = read_next_tlv(osef_stream)
        raw_tree = build_tree(tlv)
        frame_dict = parse_to_lazy_dict(raw_tree) if lazy else parse_to_dict(raw_tree)
        return frame_dict
    except (MalformedTlvException, EOFError, OsefConnectionException) as exp:
        raise OsefParsingException from exp


def _sleep(
    first_frame_counter: float,
    first_frame_timestamp: float,
    timestamp_microsecond: int,
):
    """Function to compute the pause duration required to process the parser in the frequency of the osef file"""
    current_frame_counter = time.perf_counter()
    pause = (timestamp_microsecond - first_frame_timestamp) - (
        current_frame_counter - first_frame_counter
    )

    if pause > 0.0:
        _sleep_ms(pause * 1000)


def _sleep_ms(duration: float):
    """Sleep for a precise duration.

    :param duration: Duration to sleep in seconds.
    """
    end_time = time.perf_counter_ns() + duration * 1000000
    while time.perf_counter_ns() < end_time:
        time.sleep(5e-4)



"""Module to define OSEF lib typing."""

# Third party imports
import numpy as np

# Project imports
from osef.spec import constants


class BinaryValueFormat:
    """Class to define the binary format for reading values."""

    NB_POINTS = "<L"
    NB_OBJECTS = "<L"
    NB_LAYERS = "<L"
    LIDAR_MODEL = "<B"
    DIVERGENCE = "<f"
    COORDINATES_REFERENCE_SYSTEM = "<B"
    TIME_DOMAIN = "<L"


# Define format for dictionary values.
# First value is the binary format, then the dict keys.
ImageDimensionDictFormat = (
    "<LL",
    [constants.ExtraOsefKeys.IMAGE_WIDTH, constants.ExtraOsefKeys.IMAGE_HEIGHT],
)

RangeAzimuthDictFormat = (
    "<ff",
    [
        constants.ExtraOsefKeys.AZIMUTH_BEGIN_DEG,
        constants.ExtraOsefKeys.AZIMUTH_END_DEG,
    ],
)

TimestampLidarVelodyneDictFormat = (
    "<LL",
    [constants.ExtraTimestampKeys.UNIX_S, constants.ExtraTimestampKeys.REMAINING_US],
)

GravityDictFormat = (
    "<fff",
    [
        constants.ExtraCoordinatesKeys.X,
        constants.ExtraCoordinatesKeys.Y,
        constants.ExtraCoordinatesKeys.Z,
    ],
)

# Define format for arrays.
# (typing, number of values)
PoseArrayFormat = (np.dtype(np.float32), 12)


SphericalCoodinatesDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.AZIMUTH, np.float32),
            (constants.ExtraCoordinatesKeys.ELEVATION, np.float32),
            (constants.ExtraCoordinatesKeys.DISTANCE, np.float32),
        ]
    ),
)

# 4th coordinate is not used.
SphericalCoodinates4fDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.AZIMUTH, np.float32),
            (constants.ExtraCoordinatesKeys.ELEVATION, np.float32),
            (constants.ExtraCoordinatesKeys.DISTANCE, np.float32),
            (constants.ExtraOsefKeys._TO_DROP, np.float32),
        ]
    ),
)

BoundingBoxesDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.X_MIN, np.float32),
            (constants.ExtraCoordinatesKeys.Y_MIN, np.float32),
            (constants.ExtraCoordinatesKeys.X_MAX, np.float32),
            (constants.ExtraCoordinatesKeys.Y_MAX, np.float32),
        ]
    ),
)

# 4th coordinate is not used.
CartesianCoordinates4fDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.X, np.float32),
            (constants.ExtraCoordinatesKeys.Y, np.float32),
            (constants.ExtraCoordinatesKeys.Z, np.float32),
            (constants.ExtraOsefKeys._TO_DROP, np.float32),
        ]
    ),
)

ZoneObjectBindingDType = np.dtype(
    (
        [
            (constants.ExtraZoneBindingKeys.OBJECT_ID, np.uint64),
            (constants.ExtraZoneBindingKeys.ZONE_INDEX, np.uint32),
        ]
    ),
)

ZoneObjectBinding32bDType = np.dtype(
    (
        [
            (constants.ExtraZoneBindingKeys.OBJECT_ID, np.uint32),
            (constants.ExtraZoneBindingKeys.ZONE_INDEX, np.uint32),
        ]
    ),
)

GeographicPoseDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.LATITUDE, np.float32),
            (constants.ExtraCoordinatesKeys.LONGITUDE, np.float32),
            (constants.ExtraCoordinatesKeys.HEADING, np.float32),
        ]
    )
)

GeographicPosePreciseDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.LATITUDE, np.float64),
            (constants.ExtraCoordinatesKeys.LONGITUDE, np.float64),
            (constants.ExtraCoordinatesKeys.HEADING, np.float32),
        ]
    )
)

GeographicSpeedDType = np.dtype(
    (
        [
            (constants.ExtraCoordinatesKeys.SPEED, np.float32),
            (constants.ExtraCoordinatesKeys.HEADING, np.float32),
        ]
    )
)


BackgroundSceneParamsDType = np.dtype(
    (
        [
            (constants.ExtraBackgroundKeys.WIDTH, np.uint32),
            (constants.ExtraBackgroundKeys.HEIGHT, np.uint32),
            (constants.ExtraBackgroundKeys.FIRST_AZIMUTH, np.float32),
            (constants.ExtraBackgroundKeys.AZIMUTH_STEP, np.float32),
        ]
    )
)

BackgroundSceneFragmentInfoDType = np.dtype(
    (
        [
            (constants.ExtraBackgroundKeys.FIRST_INDEX, np.uint32),
            (constants.ExtraBackgroundKeys.CELLS_NUMBER, np.uint32),
        ]
    )
)
